import sys, logging

from aug_sfutils import libddc

logger = logging.getLogger('aug_sfutils.getlastshot')
date_fmt = '%Y-%m-%d'
#logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


def parse_url(url, blen=100):
    """
    Read URL content
    """

    try:
        import urllib3
        url_lib = 'urllib3'
    except:
        try:
            import urllib2
            url_lib = 'urllib2'
        except:
            try:
                import urllib
                url_lib = 'urllib'
            except:
                logger.error('No urllib(2, 3) found')
                return None

    url_not_found = 'URLs %s not found' %url
    if url_lib == 'urllib3':
        http = urllib3.PoolManager()
        try:
            tmp = http.request('GET', url)
        except:
            logger.error(url_not_found)
            return None
        bshot = tmp.data[:blen]
    elif url_lib == 'urllib2':
        try:
            bshot = urllib2.urlopen(url).read(blen)
        except:
            logger.error(url_not_found)
            return None
    elif url_lib == 'urllib':
        try:
            bshot = request.urlopen(url).read()
        except:
            logger.error(url_not_found)
            return None

    return bshot


def getlastshot(method=None):
    """
    Gets last shot number from the AUG webpage(s)
    Optional: method ('jou')
    """

    if method is None:
        logger.debug('Using JOU shotfile')
        nshot = libddc.previousshot('JOU', 99999)
        if nshot is not None:
            return nshot

    logger.debug('Using URL ShotNumber.dta')
    url1 = 'http://ssr-mir.aug.ipp.mpg.de:9090/Diag/ShotNumber.dta'
    nshot = parse_url(url1, blen=5)

    if nshot is None:
        logger.error('Cannot convert None to integer')
    else:
        nshot = int(nshot)
 
    return nshot



import numpy as np


class SFOBJ(np.ndarray):
    """
    Subclass of numpy.ndarray adding a few metadata,
    such as physical unit, description and relations
    """

    def __new__(cls, input_array, sfho=None):

        obj = np.asarray(input_array).view(cls)
        for attr, val in sfho.__dict__.items():
            obj.__dict__[attr] = val
        return obj

    def __array_finalize__(self, obj):

        if obj is None: return
        for attr in ('objnam', 'obj_type', 'object_type', 'level', 'status', 'errcode', 'rel', 'relations', 'address', 'length', 'descr', 'data_format', 'physunit', 'phys_unit', 'stat_ext', 'n_pre', 'n_steps', 'tbase_type'):
            self.__dict__[attr] = getattr(obj, attr, None)



import sys
import numpy as np

def to_byt(str_or_byt):
    """
    Converts a plain string to a byte string (python2/3 compatibility)
    """

    str_out = str_or_byt.encode('utf8') if not isinstance(str_or_byt, bytes) else str_or_byt

    return str_out


def to_str(str_or_byt, strip=True):
    """
    Converts a byte string to a plain string (python2/3 compatibility)
    """

    if sys.version_info >= (3, 0, 0):
        if isinstance(str_or_byt, bytes):
#            str_out = str_or_byt.decode('utf8')
            str_out = str_or_byt.decode('iso-8859-1').replace('\x00', '')
        else:
            str_out = str_or_byt
    else:
        if isinstance(str_or_byt, unicode):
            str_out = str(str_or_byt)
        else:
            str_out = str_or_byt

    if strip:
        return str_out.strip()
    else:
        return str_out


def to_small_endian(var):

    var = np.array(var)
    if var.dtype.byteorder == '>':
        var = var.newbyteorder().byteswap(inplace=True)
    return var


def next8(nlen):

    return 8 * ((nlen + 7) // 8)



"""Editing shotfile headers"""

import shutil, os, logging
import ctypes as ct
import numpy as np
from aug_sfutils import sfdics, str_byt, sfread

sfhlib = '/afs/ipp/aug/ads/lib64/@sys/libsfh8.so'
if not os.path.isfile(sfhlib):
    sfhlib = '/afs/ipp/aug/ads/lib64/amd64_sles11/libsfh8.so'

libsfh = ct.cdll.LoadLibrary(sfhlib)

logger = logging.getLogger('aug_sfutils.sfh')

for key, val in sfdics.typ2descr.items():
    if key is not None:
        exec('%s=%d' %(val, key))


class SFH:

    def Open(self, fname):
        """Opens the shot file header name, reads it in a temporary memory buffer and records the last modification date."""

        fname = str_byt.to_byt(fname)
        c_name = ct.c_char_p(fname)
        self.c_sfid = ct.c_int32(0)
        _sfid = ct.byref(self.c_sfid)

        self.sfhead = sfread.SFREAD(sfh=fname)
        err = libsfh.sfhopen(c_name, _sfid, len(fname))
        libsfh.sfherror(err, 'Open')

        return err

    def Close(self):
        """This routine does the following things:
- checks if sfid is still valid
- checks if the original shot file header is unchanged, otherwise there will be an error
- copies the original shot file header to name.BAK
- writes the new shot file header."""

        err = libsfh.sfhclose(self.c_sfid)
        libsfh.sfherror(err, 'Close')

        return err

    def Mdarea(self, obj, nsteps, nx, ny, nz):
        """Modifes the size of an AreaBase object
        """

        bobj  = str_byt.to_byt(obj)
        c_obj = ct.c_char_p(bobj)
        c_nsteps = ct.c_uint32(nsteps)
        c_nx = ct.c_uint32(nx)
        c_ny = ct.c_uint32(ny)
        c_nz = ct.c_uint32(nz)

        err = libsfh.sfhmdarea(self.c_sfid, c_obj, c_nsteps, c_nx, c_ny, c_nz)
        libsfh.sfherror(err, 'mdarea')

    def Mdindex(self, obj, nx, ny, nz, qual=False):
        """Modifes the spatial index of an object
        """

        bobj  = str_byt.to_byt(obj)
        c_obj = ct.c_char_p(bobj)
        c_nx  = ct.c_uint32(nx)
        c_ny  = ct.c_uint32(ny)
        c_nz  = ct.c_uint32(nz)

        if qual:
            err = libsfh.sfhmdqualindex(self.c_sfid, c_obj, c_nx, c_ny, c_nz)
        else:
            err = libsfh.sfhmdindex24(self.c_sfid, c_obj, c_nx, c_ny, c_nz)
        libsfh.sfherror(err, 'mdindex24')

    def Modtim(self, obj, nt):
        """Modifes the size of the time dimension of an object
        """

        bobj  = str_byt.to_byt(obj)
        c_obj = ct.c_char_p(bobj)
        c_nt  = ct.c_uint32(nt)

        err = libsfh.sfhmodtim(self.c_sfid, c_obj, c_nt)
        libsfh.sfherror(err, 'modtim')

    def Modindex1(self, obj, nt):

        bobj  = str_byt.to_byt(obj)
        c_obj = ct.c_char_p(bobj)
        c_nt  = ct.c_uint32(nt)

        err = libsfh.sfhmdindex1(self.c_sfid, c_obj, c_nt)
        libsfh.sfherror(err, 'modindex1')

    def Modsgr(self, obj, dims, qual=False):
        """Modifes the dimenskions of a SigGroup
        """

        nx = 1 + np.zeros(4, dtype=int)
        for jdim in range(len(dims)):
            nx[jdim] = dims[jdim]
        self.Modtim(obj, nx[0])
        self.Mdindex(obj, nx[1], nx[2], nx[3], qual=qual)

    def Modpar(self, ps, pn, dat):
        """Modifies a parameter's size in ParSet
        """

        pss = str_byt.to_str(ps)
        pns = str_byt.to_str(pn)
        if pss not in self.sfhead.parsets:
            logger.debug('%s it not a parameter set', pss)
            return

        dfmt = self.sfhead(pss)[pn].data_format
        if dfmt in sfdics.fmt2len.keys():
            sf_dtype = CHAR
        else:
            sf_dtype = sfdics.fmt2typ[dfmt]
        pnlen = len(dat)
# Input
        bps   = str_byt.to_byt(ps)
        bpn   = str_byt.to_byt(pn)
        bdat  = str_byt.to_byt(dat)
        c_ps  = ct.c_char_p(bps)
        c_pn  = ct.c_char_p(bpn)
        c_typ = ct.c_uint32(sf_dtype)
        c_len = ct.c_uint32(pnlen)
        _typ  = ct.byref(c_typ)
        _len  = ct.byref(c_len)
# Output

        logger.debug('Modpar test: %s %d %d %d', pns, sf_dtype, pnlen, len(dat))
        if dfmt in sfdics.fmt2len.keys():
            name_len = sfdics.fmt2len[dfmt]
            c_len  = ct.c_uint32(pnlen)
            c_data = ct.c_char_p(bdat)
            err = libsfh.sfhmodpar(self.c_sfid, c_ps, c_pn, c_typ, c_len, c_data)
        else:
            err = libsfh.sfhmodpar(self.c_sfid, c_ps, c_pn, c_typ, c_len, dat.ctypes.data_as(ct.POINTER(ct.c_long)))
        libsfh.sfherror(err, 'Modpar')


    def newobj(self, objname, objtyp, subtyp):
        """Creates a new SFH object
        """

# objtyp: 6 SGR, 7 Signal, 8 Timebase, 13 Areabase
        objname = str_byt.to_byt(objname)
        c_obj = ct.c_char_p(objname)
        c_otyp = ct.c_uint16(objtyp)
        c_styp = ct.c_uint16(subtyp)
        err = libsfh.sfhnewobj(self.c_sfid, c_obj, c_otyp, c_styp)
        libsfh.sfherror(err, 'newobj')


    def newrel(self, objname, newrelname):
        """Adds a new relation to a given object
        """

        objname    = str_byt.to_byt(objname)
        newrelname = str_byt.to_byt(newrelname)
        c_obj = ct.c_char_p(objname)
        c_rel = ct.c_char_p(newrelname)
        err = libsfh.sfhstrel(self.c_sfid, c_obj, c_rel)
        libsfh.sfherror(err, 'newrel')


    def newrelt(self, objname, newreltb):
        """Adds a new TimeBase relation to a given object
        """

        objname  = str_byt.to_byt(objname)
        newreltb = str_byt.to_byt(newreltb)
        c_obj   = ct.c_char_p(objname)
        c_reltb = ct.c_char_p(newreltb)
        err = libsfh.sfhstreltb(self.c_sfid, c_obj, c_reltb)
        libsfh.sfherror(err, 'newrelt')


    def set_text(self, objname, newtext):
        """Replaces the description string of a given object
        """

        objname = str_byt.to_byt(objname)
        newtext = str_byt.to_byt(newtext)
        c_obj  = ct.c_char_p(objname)
        c_text = ct.c_char_p(newtext)
        err = libsfh.sfhmtext(self.c_sfid, c_obj, c_text)
        libsfh.sfherror(err, 'set_text')

#===============
#    New Methods
#===============

    def Mapping(self):
# Bug: use sfhread!
        objects = self.Lonam()
        if objects is None:
            return
        mapping = []

        for i in objects:
            if objects[i] == 6:
                indices = self.Rdindex24(i)
                for i2 in range(indices[2]):
                    for i3 in range(indices[3]):
                        for i4 in range(indices[4]):
                            mapping.append([i, [i2, i3, i4], \
                                self.Rdmap(i, i2 + 1 ,i3 + 1, i4 + 1)['chan']])

        return mapping

    def Channelmapping(self):
    
        raw_mapping = self.Mapping()
        mapping_array = np.array(len(raw_mapping), dtype=np.str)
        for raw_map in raw_mapping:
            mapping_array[raw_map[2]] = raw_map[0] + '_' + str(raw_map[1][0]) + '-' + str(raw_map[1][1]) + '-' + str(raw_map[1][2])

        return mapping_array

    def Signalmapping(self):
    
        raw_mapping = self.Mapping()
        mapping_table = {}
        for raw_map in raw_mapping:
            mapping_table[raw_map[0] + '_' + str(raw_map[1][0]) + '-' + str(raw_map[1][1]) + '-' + str(raw_map[1][2])] = raw_map[2]

        return mapping_table



"""Equilibrium object from shotfile"""

import logging
import numpy as np
from scipy.integrate import cumtrapz

from aug_sfutils import sfobj, SFREAD, parse_kwargs

logger = logging.getLogger('aug_sfutils.sf2equ')

#--------------------------------
# COCOs
# https://crppwww.epfl.ch/~sauter/cocos/Sauter_COCOS_Tokamak_Coordinate_Conventions.pdf
# Table I page 8
#--------------------------------

sigma = { \
    'bp'       : [ 1,  1, -1, -1,  1,  1, -1, -1], \
    'rphiz'    : [ 1, -1,  1, -1,  1, -1,  1, -1], \
    'rhothephi': [ 1,  1, -1, -1, -1, -1,  1,  1]}

explain = {'rphiz'    : {1: '(R, phi, Z) r'    , -1: '(R, Z, phi) l'}, \
           'rhothephi': {1: '(rho, the, phi) r', -1: '(rho, phi, the) l'    } }


class EQU:
    """Reads the whole equilibrium shotfile
    Reduces 1D profiles to within separatrix,
    Calculates a few useful derived quantities
    Accepts a time sub-interval via tbeg, tend keywords
    """

    def __init__(self, nshot, **kwargs):

        """
        Input
        ---------
        nshot: int
            shot number
        diag: str
            diagnsotics used for mapping (EQI, EQH, ...)
        exp: str
            experiment (AUGD)
        ed:  int
            edition
        tbeg: float
            initial time [s]
        tend: float
            final time [s]
        """

        exp  = parse_kwargs.parse_kw( ('exp', 'experiment'), kwargs, default='AUGD')
        diag = parse_kwargs.parse_kw( ('diag', 'diagnostic'), kwargs, default='EQH')
        ed   = parse_kwargs.parse_kw( ('ed', 'edition'), kwargs, default=0)

        self.sf = SFREAD(nshot, diag, ed=ed, exp=exp)

        if self.sf.status:

            ## Shot number
            self.shot = nshot 

            ## Shotfile exp
            self.exp = exp
            
            ## Shotfile diag
            self.diag = diag

            ## Actual shotfile edition (> 0)
            self.ed = self.sf.ed

            ## COCO number
            self.cocos = 17

            ## Time grid of equilibrium shotfile
            self.time = self.sf.getobject('time')
            nt = len(self.time)
            parmv = self.sf.sfh['PARMV'].pars
            nR = parmv['M'] + 1
            nZ = parmv['N'] + 1
            ## R of PFM cartesian grid [m]
            self.Rmesh = self.sf.getobject('Ri')[: nR, 0]
            ## z of PFM cartesian grid [m]
            self.Zmesh = self.sf.getobject('Zj')[: nZ, 0]

            Lpf = self.sf.getobject('Lpf').T
            ## For 1D profiles: #points within separatrix
            self.lpfp = Lpf%10000 # inside sep
            ## For 1D profiles: #points in SOL
            self.lpfe = (Lpf - self.lpfp)//10000 #outside sep
            ## Spatial index sorted from inside to outside, at each time
            self.ind_sort = []
            for jt in range(nt):
                ind_core = np.arange(self.lpfp[jt], -1, -1)
                ind_sol  = np.arange(self.lpfp[jt] + 3, self.lpfp[jt] + 3 + self.lpfe[jt])
                self.ind_sort.append(np.append(ind_core, ind_sol))
        else:
            logger.error('Problems opening %s:%s:%d for shot %d' %(exp, diag, ed, nshot))
            return

        if 'tbeg' in kwargs.keys():
            tbeg = kwargs['tbeg']
        else:
            tbeg = np.min(self.time)
        if 'tend' in kwargs.keys():
            tend = kwargs['tend']
        else:
            tend = np.max(self.time)

        (self.indt, ) = np.where((self.time >= tbeg) & (self.time <= tend))
        if len(self.indt) == 0:
            self.indt = [np.argmin(np.abs(self.time - tbeg))]

        logger.info('Reading equ scalars')
        self.read_scalars()
        logger.info('Reading equ 1d profiles')
        self.read_profiles()
        logger.info('Reading equ PFM')
        self.read_pfm()

        logger.info('COCO %d' %self.cocos)

# Sign of Ip, Bt (positive=ccw from above), valid for COCO=17 (CLISTE)
        self.ip_ccw = np.sign((self.psix - self.psi0).mean())
        self.bt_ccw = np.sign(np.nanmean(self.jpol))

        jcoco = self.cocos%10 - 1
        ebp   = self.cocos//10
        self.psi_sign = -self.ip_ccw*sigma['bp'][jcoco]*sigma['rphiz'][jcoco]
        psi_2pi  = (2.*np.pi)**ebp
        self.psi_fac = self.psi_sign*psi_2pi
        self.phi_sign = self.bt_ccw*sigma['rphiz'][jcoco]

        self.time = self.time[self.indt]
        nt = len(self.time)
        nrho = np.max(self.lpfp) + 1

        self.psi0 = self.psi0[self.indt]
        self.psix = self.psix[self.indt]
        if hasattr(self, 'ipipsi'):
            self.ipipsi = self.ipipsi[self.indt]

        for key in self.ssqnames:
            self.__dict__[key] = self.__dict__[key][self.indt]
        for key in ('pfl', 'tfl', 'ffp', 'rinv', 'q', \
             'area',  'vol',  'pres',  'jpol', \
            'darea', 'dvol', 'dpres', 'djpol') :
            if self.__dict__[key] is not None:
                self.__dict__[key] = self.__dict__[key][self.indt, :nrho]

        ## Geometric major axis [m]
        self.R0 = 1.65
        ## Central magnetic field F/R [T]
        self.B0 = self.jpol[..., 0]*2e-7/self.R0
        ## Magnetic field on magnetic axxis [T]
        self.Baxis = self.jpol[..., 0]*2e-7/self.Rmag
        ## Normalised poloidal flux [Vs]
        self.psiN = (self.pfl             - self.psi0[..., None]) / \
                    (self.psix[..., None] - self.psi0[..., None])

        grad_area = np.apply_along_axis(np.gradient, -1, self.area)

        if self.ffp is not None and self.rinv is not None:
            self.jav = 2.*np.pi*self.dpres/self.rinv + self.ffp*self.rinv/2e-7
            curr_prof = np.cumsum(self.jav*grad_area, axis=-1)
        ## Plasma current [A] as integral of current density
            self.ip = curr_prof[..., -1]

        self.q[..., -1] = 2*self.q[..., -2] - self.q[..., -3]
        ## Unnormalised rho toroidal
        self.rho_tor = np.sqrt(np.abs(self.tfl/(np.pi*self.B0[..., None])))
        ## Normalised rho toroidal
        self.rho_tor_n = np.zeros_like(self.tfl)
        self.rhotor_trapz = np.zeros_like(self.tfl) # for RABBIT
        ## Normalised toroidal flux [Vs]
        self.tfln = np.zeros_like(self.tfl)
        for jt in range(nt):
            tf = self.tfl[jt]
            torflux_trapz = cumtrapz(-self.q[jt], x=self.pfl[jt], initial=0.)
            self.tfln[jt, :] = tf/tf[-1]
            self.rho_tor_n[jt, 1:] = np.sqrt((tf[1:] - tf[0])/(tf[-1] - tf[0]))
            self.rhotor_trapz[jt, 1:] = np.sqrt( torflux_trapz[1:] / (torflux_trapz[-1] - torflux_trapz[0]) )

        self.pfm = self.pfm[:, :, self.indt]


    def read_pfm(self):

        """
        Reads PFM matrix
        """

        nt = len(self.time)
        nR = len(self.Rmesh)
        nZ = len(self.Zmesh)
        ## Poloidal flux matrix [Vs]
        self.pfm = self.sf.getobject('PFM')[: nR, : nZ, :nt]


    def read_ssq(self):

        """
        Creating attributes corresponding to SSQNAM.
        Beware: in different shotfiles, for a given j0 
        SSQ[j0, time] can represent a different variable.
        This routine handles it consistently.
        """

        ssqs   = self.sf.getobject('SSQ') # Time is second
        ssqnam = self.sf.getobject('SSQnam')

        nt = len(self.time)
        ## Names of SSQ parameters
        self.ssqnames = []
        for jssq in range(ssqnam.shape[1]):
            tmp = b''.join(ssqnam[:, jssq]).strip()
            lbl = tmp.decode('utf8')
            if lbl.strip() != '':
                if lbl not in self.ssqnames: # avoid double names
                    self.ssqnames.append(lbl)
                    ## Scalar quantities
                    self.__dict__[lbl] = ssqs[jssq, :nt]


    def read_scalars(self):

        """
        Reads R, z, psi at magnetic axis and separatrix, only if attribute 'r0' is missing.
        equ_map.r0 is deleted at any equ_map.Open, equ_map.Close call.
        """

        self.read_ssq()

        nt = len(self.time)
# Position of mag axis, separatrixx

        PFxx = self.sf.getobject('PFxx')[:, :nt]
        ikCAT = np.argmin(abs(PFxx[1:, :] - PFxx[0, :]), axis=0) + 1
        if all(PFxx[2]) == 0: ikCAT[:] = 1  #troubles with TRE equilibrium 

        ## Poloidal flux at magnetic axis [Vs]
        self.psi0 = PFxx[0, :nt]
        ## Poloidal flux at separatrix [Vs]
        self.psix = PFxx[ikCAT, np.arange(nt)]

        try:
            ## Plasma current[A]
            self.ipipsi = self.sf.getobject('IpiPSI')[0]
        except:
            logger.exception('Signal IpiPSI not found')


    def read_profiles(self):
        """Reading 1D quantities, including some derivatives"""

        ## Toroidal flux(t, rho) [Vs]
        self.tfl = self.get_profile('TFLx')
        ## Poloidal flux (t, rho) [Vs]
        self.pfl = self.get_profile('PFL')
        ## Safety factor(t, rho)
        self.q     = self.get_profile('Qpsi')
        ## FF'
        self.ffp   = self.get_profile('FFP')
        self.rinv  = self.get_profile('Rinv')
        self.r2inv = self.get_profile('R2inv')
        self.bave  = self.get_profile('Bave')
        self.b2ave = self.get_profile('B2ave')

        ## dVolume/dPsi [m**3/Vs]
        self.vol , self.dvol  = self.get_mixed('Vol')
        ## dArea/dPsi [m**2/Vs]
        self.area, self.darea = self.get_mixed('Area')
        ## dPres/dPsi [Pa/Vs]
        self.pres, self.dpres = self.get_mixed('Pres')
        self.jpol, self.djpol = self.get_mixed('Jpol')


    def get_profile(self, var_name):

        """
        var_name: str
            name of the quantity, like
            Qpsi       q_value vs PFL
            Bave       <B>vac
            B2ave      <B^2>vac
            FFP        ff'
            Rinv       <1/R>
            R2inv      <1/R^2>
            FTRA       fraction of the trapped particles
        Output:
            array(t, rho) with some metadata (units, description)
        """

        nt = len(self.time)

        profs = ('TFLx', 'PFL', 'FFP', 'Qpsi', 'Rinv', 'R2inv', 'Bave', 'B2ave')

        if var_name not in profs:
            logger.error('SignalGroup %s unknown' %var_name)
            return None

        tmp = self.sf.getobject(var_name)
        if tmp is None:
            return None
        var = tmp[:, :nt]

        nrho = np.max(self.lpfp + 1 + self.lpfe)
        
        var_sorted = sfobj.SFOBJ(np.zeros((nt, nrho)), sfho=tmp)
        for jt in range(nt):
            sort_wh = self.ind_sort[jt]
            nr = len(sort_wh)
            var_sorted [jt, :nr] = var[sort_wh, jt]
            if nr < nrho:
                var_sorted[jt, nr:] = var_sorted[jt, nr-1]

        return var_sorted


    def get_mixed(self, var_name):

        """
        var_name: str
            name of the quantity, like
            Jpol       poloidal current,
            Pres       pressure
            Vol        plasma Volume
            Area       plasma Area
        Output:
            array-pair (quantity(t, rho), derivative(t, rho)) with some metadata (units, description)
        """

        nt = len(self.time)
        mixed = ('Vol', 'Area', 'Pres', 'Jpol')

# Pairs prof, d(prof)/d(psi)

        if var_name not in mixed:
            logger.error('%s not one of the mixed quanties Vol, Area, Pres, Jpol' %var_name)
            return None, None

        tmp = self.sf.getobject(var_name)
        var  = tmp[ ::2, :nt]
        dvar = tmp[1::2, :nt]

        nrho = np.max(self.lpfp + 1 + self.lpfe)
        info  = type('', (), {})()
        dinfo = type('', (), {})()
        for key, val in tmp.__dict__.items():
            if key != 'data':
                info.__dict__[key] = val
                if key == 'phys_unit':
                    dinfo.__dict__[key] = '%s/%s' %(val, self.pfl.phys_unit)
                else:
                    dinfo.__dict__[key] = val
        var_sorted  = sfobj.SFOBJ( np.zeros((nt, nrho)), sfho=info)
        dvar_sorted = sfobj.SFOBJ( np.zeros((nt, nrho)), sfho=dinfo)
        for jt in range(nt):
            sort_wh = self.ind_sort[jt]
            nr = len(sort_wh)
            var_sorted [jt, :nr] = var [sort_wh, jt]
            dvar_sorted[jt, :nr] = dvar[sort_wh, jt]
            if nr < nrho:
                var_sorted [jt, nr:] = var_sorted[jt, nr-1]

        return var_sorted, dvar_sorted


    def find_coco(self, ip_shot='ccw', bt_shot='cw'):
        """
        Identifies the COCO number of a given equilibrium object
        """

# dpsi_sign positive if psi_sep > psi0
        dpsi_sign = np.sign(np.mean(self.psix) - np.mean(self.psi0))
# Known plasma discharge
        ccw_ip = 1 if(ip_shot == 'ccw') else -1 # AUG default: 1
        ccw_bt = 1 if(bt_shot == 'ccw') else -1 # AUG default: -1

# Table III

        sign_q  = np.sign(np.nanmean(self.q))
        sign_ip = np.sign(np.nanmean(self.ipipsi))
        sign_bt = np.sign(np.nanmean(self.jpol))
        sigma_rphiz = sign_ip*ccw_ip
        sigma_bp    = dpsi_sign*sign_ip
# Eq 45
        sigma_rhothephi = sign_q*sign_ip*sign_bt
        logger.debug(sigma_bp, sigma_rphiz, sigma_rhothephi)
        for jc, rhothephi in enumerate(sigma['rhothephi']):
            if(sigma['bp'   ][jc] == sigma_bp    and \
               sigma['rphiz'][jc] == sigma_rphiz and \
               rhothephi          == sigma_rhothephi):
                self.cocos = jc + 1
                break

# Find out 2*pi factor for Psi

        dphi = np.gradient(self.tfl, axis=1)
        dpsi = np.gradient(self.pfl, axis=1)

# Radial+time average
# It is either q_ratio ~ 1 (COCO > 10) or ~ 2*pi (COCO < 10)
        q_ratio = np.abs(np.nanmean(dphi/(self.q*dpsi)))
        logger.debug('Ratio %8.4f' %q_ratio)
        if q_ratio < 4:
            self.cocos += 10


    def to_coco(self, cocos_out=11):
        """
        Transforms equilibrium object into any wished output COCO
        """

# Assuming SI

        cocos_in = self.cocos
        logger.info('COCO conversion from %d to %d' %(cocos_in, cocos_out))
        jc_in   = cocos_in %10 - 1
        jc_out  = cocos_out%10 - 1
        ebp_in  = cocos_in//10
        ebp_out = cocos_out//10
#    sign_ip_in = np.sign(np.nanmean(self.ipipsi))
# Equation 9, table I, equation 39, 45
        q_sign   = sigma['rhothephi'][jc_in]*sigma['rhothephi'][jc_out]
        phi_sign = sigma['rphiz'][jc_in]*sigma['rphiz'][jc_out]
        psi_sign = sigma['rphiz'][jc_in]*sigma['bp'][jc_in] * sigma['rphiz'][jc_out]*sigma['bp'][jc_out]
        psi_2pi  = (2.*np.pi)**(ebp_out - ebp_in)
        psi_fac = psi_sign*psi_2pi
        try:
            logger.debug(np.mean(self.jav), phi_sign)
        except:
            pass

        for key, val in self.__dict__.items():
            if val is None:
                continue
            if key in ('B0', 'jpol', 'jav', 'tfl', 'ip', 'ipipsi', 'phi_sign'):
                self.__dict__[key] = val*phi_sign
            elif key in ('psi0', 'psix', 'pfl', 'pfm', 'psi_fac'):
                self.__dict__[key] = val*psi_fac
            elif key in ('dpres', 'darea', 'dvol', 'ffp'):
                self.__dict__[key] = val/psi_fac
            elif key in ('djpol', ):
                self.__dict__[key] = val*phi_sign/psi_fac
            elif key in ('q', 'q0', 'q25', 'q50', 'q75', 'q95'):
                self.__dict__[key] = val*q_sign
        self.cocos = cocos_out


if __name__ == '__main__':


    nshot = 28053
    eqm = EQU(nshot)
    logger.info(eqm.ssqnames)



import os, datetime, logging
from struct import unpack
import numpy as np

from aug_sfutils import sfdics, sfhread, sfobj, manage_ed, parse_kwargs, str_byt, libddc, getlastshot

PPGCLOCK = [1e-6, 1e-5, 1e-4, 1e-3]

logger = logging.getLogger('aug_sfutils.sfread')
date_fmt = '%Y-%m-%d'
#logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


for key, val in sfdics.obj_name.items():
    exec('%s=%d' %(val, key))
for key, val in sfdics.cal_type.items():
    exec('%s=%d' %(val, key))
for key, val in sfdics.fmt2descr.items():
    exec('%s=%d' %(val, key))
for key, val in sfdics.tbtype_d.items():
    exec('%s=%d' %(val, key))


def read_other_sf(*args, **kwargs):

    return SFREAD(*args, **kwargs)


def getcti_ts06(nshot):
    """
    Gets the absolute time (ns) of a discharge trigger 
    """
    diag = 'CTI'
    cti = SFREAD(nshot, diag)
    try:
        cdev = cti.getdevice('LAM')
        ts06 = cdev['PhyReset']
        if ts06 == 0:
            ts06 = cdev['TS06']
        if ts06 == 0:
            ts06 = cdev['CT_TS06']
    except: # shot < 35318
        cdev = cti.getdevice('TS6')
        ts06 = cdev['TRIGGER']
        logger.debug('getcti_ts06 %d', ts06)
    if ts06 < 1e15:
        ts06 = None
    return ts06


def getgc(nshot=30136):
    """
    Returns first wall contours for plotting
    """
    ygc_sf = [1996, 8646, 8650, 9401, 11300, 11301, 11320, 12751, 13231, 14051, 14601, 16310, 16315, 18204, 19551, 21485, 25891, 30136]
    ygc_sf = np.array(ygc_sf)
    nshot = np.max(ygc_sf[ygc_sf <= nshot])
    ygc = SFREAD(nshot, 'YGC')
    if ygc is not None:
        rrgc     = ygc.getobject('RrGC', cal=False)
        zzgc     = ygc.getobject('zzGC', cal=False)
        inxlen   = ygc.getobject('inxlen', cal=False)
        flag_use = ygc.getobject('ixplin', cal=False)
        gcnames  = ygc.getobject('chGCnm', cal=False)

    comp_d = {}

    inxbeg = np.cumsum(np.append(0, inxlen))

    if gcnames is None:
        for jcom, leng in enumerate(inxlen):
            comp_d[jcom] = type('', (), {})() # empty object
            comp_d[jcom].r = rrgc[inxbeg[jcom]: inxbeg[jcom+1]]
            comp_d[jcom].z = zzgc[inxbeg[jcom]: inxbeg[jcom+1]]
    else:
        for jcom, lbl in enumerate(gcnames):
            lbl = str_byt.to_str(lbl)
            if flag_use[jcom] > 0:
                comp_d[lbl] = type('', (), {})()
                comp_d[lbl].r = rrgc[inxbeg[jcom]: inxbeg[jcom+1]]
                comp_d[lbl].z = zzgc[inxbeg[jcom]: inxbeg[jcom+1]]

    return comp_d


class SFREAD:
    """
    Class for reading ASDEX Upgrade shotfile data
    """

    def __init__(self, *args, **kwargs):
        """
        Opens a shotfile, reads the header
        """

        self.shot = None
        self.diag = None
        self.exp = None
        self.status = False
        self.open(*args, **kwargs)
        if len(args) > 2:
            logger.warning('More than 2 explicit arguments: only the first two (diag, shot) are retained')


    def open(self, *args, **kwargs):

        if 'sfh' in kwargs.keys():
            self.sfile = kwargs['sfh']
            self.shot = 0
            self.ed = 0
            self.diag = os.path.basename(self.sfile)[:3]

        elif 'sf' in kwargs.keys():
            self.sfile = os.path.abspath(kwargs['sf'])
            dirs = self.sfile.split('/')[::-1]
            sshot = ''
            for subdir in dirs:
                try:
                    a = float(subdir)
                    sshot = subdir + sshot
                except:
                    self.diag = subdir
                    break
            self.shot = int(sshot.split('.')[0])

        else:

            n_args = len(args) 
            if n_args == 0:
                logger.warning('No argument given, need at least diag_name')
                return
            if isinstance(args[0], str) and len(args[0].strip()) == 3:
                diag = args[0].strip()
                if n_args > 1: 
                    if isinstance(args[1], (int, np.integer)):
                        nshot = args[1]
            elif isinstance(args[0], (int, np.integer)):
                nshot = args[0]
                if n_args > 1:
                    if isinstance(args[1], str) and len(args[1].strip()) == 3:
                        diag = args[1].strip()
            if 'nshot' not in locals(): 
                logger.warning('No argument is a shot number (int), taking last AUG shot')
                nshot = getlastshot.getlastshot()
            if 'diag' not in locals():
                diag = input('Please enter a diag_name (str(3), no delimiter):\n')

            exp = parse_kwargs.parse_kw( ('exp', 'experiment'), kwargs, default='AUGD')
            ed  = parse_kwargs.parse_kw( ('ed', 'edition'), kwargs, default=0)
            logger.debug('%d %s %s %d', nshot, diag, exp, ed)
            self.sfile, self.ed = manage_ed.sf_path(nshot, diag, exp=exp, ed=ed)
            if self.sfile is None:
                logger.error('Shotfile not found for %s:%s(%d) #%d', exp, diag, ed, nshot)
                return
            else:
                self.shot = nshot
                self.diag = diag.upper()
                self.exp  = exp  # unused herein, but useful docu

        logger.debug('Shotfile path: %s', self.sfile)
        if os.path.isfile(self.sfile):
            self.time = datetime.datetime.fromtimestamp(os.path.getctime(self.sfile))
            with open(self.sfile, 'rb') as f:
                byt_str = f.read(128000)
                if len(byt_str) < 128:
                    logger.error('Error: shotfile %s has < 128 bytes, ignored' %(self.sfile))
                    return
                self.sfh = sfhread.read_sfh(byt_str)
                if self.sfh is not None:
                    self.status = True
        else:
            logger.error('Shotfile %s not found' %self.sfile)
            return

        logger.info('Fetching SF %s', self.sfile)
        self.objects = [ lbl for lbl, sfo in self.sfh.items() if sfo.obj_type in [SignalGroup, Signal, TimeBase, AreaBase] ]
        self.parsets = [ lbl for lbl, sfo in self.sfh.items() if sfo.obj_type in [Device, ParamSet] ]
        if 'ADDRLEN' in self.sfh.keys():
            size_fac = sfdics.addrsizes[self.sfh['ADDRLEN'].addrlen]
            for obj in self.sfh.keys():
                if hasattr(self.sfh[obj], 'address'):
                    self.sfh[obj].address *= size_fac

        for parset in self.parsets:
            logger.debug('ParSet %s', parset)
            try:
                self.sfh[parset].pars = self.getparset(parset)
            except:
                self.sfh[parset].pars = None

        self.cache = {}


    def __call__(self, name):

        if not self.status:
            return None

        if name in self.parsets:
            return self.sfh[name].pars

        if name not in self.cache.keys():
            if name in self.objects:
                self.cache[name] = self.getobject(name)
            elif name not in self.parsets:
                logger.error('Signal %s:%s not found for shot #%d', self.diag, name, self.shot)
                return None
        return self.cache[name]


    def getchunk(self, start, length):
        """
        Reads the requested byteblock from the actual file, not the byt_str
        """
        rdata = None
        with open(self.sfile, 'rb') as f:
            f.seek(start)
            rdata = f.read(length)
        return rdata


    def gettimebase(self, obj, tbeg=None, tend=None):
        """
        Reads the timebase of a given SIG, SGR or AB
        """

        obj = str_byt.to_str(obj)
        if obj not in self.sfh.keys():
            logger.error('Sig/TB %s:%s not found for #%d', self.diag, obj, self.shot)
            return None
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        if otyp == TimeBase:
            return self.getobject(obj, tbeg=tbeg, tend=tend)
        elif otyp in (SignalGroup, Signal, AreaBase):
            for rel in sfo.relations:
                if self.sfh[rel].obj_type == TimeBase:
                    return self.getobject(rel, tbeg=tbeg, tend=tend)
        return None


    def getareabase(self, obj, tbeg=None, tend=None):
        """
        Reads the areabase of a given SIG or SGR
        """

        obj = str_byt.to_str(obj)
        if obj not in self.sfh.keys():
            logger.error('Sig/AB %s:%s not found for #%d', self.diag, obj, self.shot)
            return None
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        if otyp == AreaBase:
            return self.getobject(obj, tbeg=tbeg, tend=tend)
        elif otyp in (SignalGroup, Signal):
            for rel in sfo.relations:
                if self.sfh[rel].obj_type == AreaBase:
                    return self.getobject(rel, tbeg=tbeg, tend=tend)
        return None


    def getobject(self, obj, cal=True, nbeg=None, nend=None, tbeg=None, tend=None):
        """
        Reads the data of a given TB, AB, SIG or SGR
        """

        obj = str_byt.to_str(obj)
        data = None
        if obj not in self.sfh.keys():
            logger.error('Signal %s:%s not found for #%d', self.diag, obj, self.shot)
            return None

# Keep commented, to allow 1 cal and 1 uncal reading
#        if obj in self.cache.keys(): 
#            return cache[obj]

        sfo = self.sfh[obj]
        if sfo.status != 0:
            logger.error('Status of SF object %s is %d' %(obj, sfo.status))
            return None
        otyp = sfo.obj_type

        if tbeg is not None or tend is not None:
            if tend is None:
                tend = 10.
            if tbeg is None:
                tbeg = 0.
            tb = self.gettimebase(obj)
            if tb is None: # AB with no time dependence
                logger.warning('%s has no Timebase associated, returning full array', obj)
                return self.getobject(obj)
            jt_beg, jt_end = tb.searchsorted((tbeg, tend))
            if otyp == TimeBase:
                return tb[jt_beg: jt_end]
            elif otyp in (Signal, AreaBase) or self.time_last(obj):
                return self.getobject(obj, cal=cal, nbeg=jt_beg, nend=jt_end)
            elif self.time_first(obj):
                return self.getobject(obj, cal=cal)[jt_beg: jt_end]
            else:
                logger.error('Object %s: tbeg, tend keywords supported only when time is first or last dim', obj)
                return None

        if otyp in (SignalGroup, Signal):
            shape_arr = np.array(sfo.index[::-1][:sfo.num_dims])
        elif otyp == TimeBase:
            shape_arr = np.array([sfo.n_steps])
        elif otyp == AreaBase:
            shape_arr = np.array([sfo.size_x + sfo.size_y + sfo.size_z, sfo.n_steps])
            logger.debug('%d %d', *shape_arr)
        else:
            logger.error('Object %s is no signal, signalgroup, timebase nor areabase, skipping')
            return None

        dfmt = sfo.data_format
        logger.debug('TB %s %d %d %d', obj, otyp, sfo.length, dfmt)
        if otyp == TimeBase and sfo.length == 0:
            if sfo.tbase_type == PPG_prog: # e.g. END:T-LM_END
                data = self.ppg_time(obj)
            else:   # ADC_intern, e.g. DCN:T-ADC-SL
                data = (np.arange(sfo.n_steps, dtype=np.float32) - sfo.n_pre)/sfo.s_rate
            dout = sfobj.SFOBJ(data, sfho=sfo) # Add metadata
        else:
            if dfmt in sfdics.fmt2len.keys(): # char variable
                dlen = sfdics.fmt2len[dfmt]
                bytlen = np.prod(shape_arr) * dlen
                data = np.chararray(shape_arr, itemsize=dlen, buffer=self.getchunk(sfo.address, bytlen), order='F')
            else: # numerical variable
                sfmt = sfdics.fmt2struc[dfmt]
                type_len = np.dtype(sfmt).itemsize
                addr = sfo.address
                bytlen = np.prod(shape_arr) * type_len
                if otyp in (Signal, TimeBase, AreaBase) or self.time_last(obj):
                    if nbeg is None:
                        nbeg = 0
                    else:
                        byt_beg = nbeg*np.prod(shape_arr[:-1])
                        addr   += byt_beg*type_len
                        bytlen -= byt_beg*type_len
                        shape_arr[-1] -= nbeg
                    if nend is not None:
                        bytlen = (nend - nbeg)*np.prod(shape_arr[:-1])*type_len
                        shape_arr[-1] = nend - nbeg

                data = np.ndarray(shape_arr, '>%s' %sfmt, self.getchunk(addr, bytlen), order='F')

# LongLong in [ns] and no zero at TS06
            if otyp == TimeBase and dfmt == LONGLONG: # RMC:TIME-AD0, SXS:Time
                logger.debug('Before getts06 dfmt:%d addr:%d len:%d data1:%d, %d', dfmt, addr, bytlen, data[0], data[1])
                data = 1e-9*(data - self.getts06(obj))
                logger.debug('%d',  self.getts06(obj))

            dout = sfobj.SFOBJ(data, sfho=sfo) # Add metadata
            dout.calib = False
# Calibrated signals and signal groups
            if otyp in (SignalGroup, Signal):
                if cal:
                    dout = self.raw2calib(dout)
                    if self.diag in ('DCN', 'DCK', 'DCR'):
                        dout.phys_unit = '1/m^2'

        return dout


    def ppg_time(self, tbobj): # Bug MAG:27204
        """
        Returns the time-array in [s] for TB of type PPG_prog
        """

        for rel in self.sfh[tbobj].relations:
            if self.sfh[rel].obj_type == Device:
                ppg = self.getdevice(rel)
                npt = self.sfh[tbobj].n_pre
                nsteps = self.sfh[tbobj].n_steps
                if not 'PRETRIG' in ppg.keys():
                    continue
                if npt > 0:
                    if ppg['PRETRIG'] > 0:
                        dt = ppg['RESOLUT'][15] * PPGCLOCK[ppg['RESFACT'][15]] + 1e-6
                    else:
                        dt = 0.
                    start_time = -dt*npt #ppg['PULSES'][0]
                else:
                    start_time = 0.
                dtyp = sfdics.fmt2np[self.sfh[tbobj].data_format]
                start_phase = start_time
                if npt > 0:
                    time_ppg = dt*np.arange(npt, dtype=dtyp) + start_phase
                    start_phase = time_ppg[-1] + dt
                else:
                    time_ppg = []
                for jphase in range(16):
                    if ppg['PULSES'][jphase] > 0:
                        dt = ppg['RESOLUT'][jphase]*PPGCLOCK[ppg['RESFACT'][jphase]]
                        tb_phase = dt*np.arange(ppg['PULSES'][jphase], dtype=dtyp) + start_phase
                        time_ppg = np.append(time_ppg, tb_phase)
                        start_phase = time_ppg[-1] + dt
                        logger.debug('Start time %d %d %d %d %.4f %.4f', jphase, ppg['PULSES'][jphase], ppg['RESOLUT'][jphase], ppg['RESFACT'][jphase], dt, start_phase)
                if len(time_ppg) == 0:
                    return None
                else:
                    return time_ppg[:nsteps]
        return None


    def raw2calib(self, data):
        """
        Calibrates an uncalibrated signal or SignalGroup
        """
# Calibrated signals and signal groups
        obj = str_byt.to_str(data.objnam)
        if data.obj_type not in (SignalGroup, Signal):
            logging.error('Calibration failed for %s: no Sig, no SGR', obj)
            return data

        pscal = self.lincalib(obj)
        if pscal is None:
            if data.phys_unit == 'counts':
                cal_fac = 1.
                for rel in data.relations:
                    if self.sfh[rel].obj_type == TimeBase:
                        cal_fac = self.sfh[rel].s_rate
                        break
                return cal_fac*np.float32(data)
            else:
                return data

        for j in range(10):
            mult = 'MULTIA0%d' %j
            shif = 'SHIFTB0%d' %j
            if mult in pscal.keys():
# we need to fix the content of pscal for signagroups
# assuming first entry wins
                if j == 0:
                    dout = data*1. # Creates a copy of a read-only array, only once
                    dout.calib = True
                multi = np.atleast_1d(pscal[mult])
                shift = np.atleast_1d(pscal[shif])
                if data.obj_type == Signal or len(multi) == 1:
                    dout *= multi[0] # MXR
                    dout += shift[0]
                else:
                    n_pars = dout.shape[1]
                    if n_pars != len(multi):
                        logger.warning('Inconsitent sizes in calibration PSet %s', obj)
                    if n_pars <= len(multi):
                        dout *= multi[: n_pars] # BLB
                        dout += shift[: n_pars]
                    else:
                        dout *= multi[0]
                        dout += shift[0]
            else:
                break
        if 'dout' in locals():
            return dout
        else:
            return data


    def lincalib(self, obj):
        """
        Returns coefficients for signal(group) calibration
        """
        obj = str_byt.to_str(obj)
        for rel in self.sfh[obj].relations:
            if self.sfh[rel].obj_type == ParamSet:
                caltyp = self.sfh[rel].cal_type
                if caltyp == LinCalib:
                    logger.info('PSet for calib: %s' %rel)
                    return self.sfh[rel].pars
                elif caltyp == extCalib:
                    diag_ext = self.sfh[rel].pars['DIAGNAME']
                    diag_ext = ''.join([str_byt.to_str(x) for x in diag_ext])
                    shot_ext = libddc.previousshot(diag_ext, shot=self.shot)
                    ext = read_other_sf(shot_ext, diag_ext)
                    return ext.sfh[rel].pars # same name in external shotfile
        return None


    def getparset(self, pset):
        """
        Returns data and metadata of a Parameter Set
        """

        pset = str_byt.to_str(pset)
        sfo = self.sfh[pset]
        otyp = sfo.obj_type
        logger.debug('PSET %s, oytp %d', pset, otyp)
        if otyp not in (Device, ParamSet):
            return None

        buf = self.getchunk(sfo.address, sfo.length)

        j0 = 0
        par_d = {}
        logger.debug('Diag: %s, PS: %s, addr: %d, length: %d', self.diag, pset, sfo.nitems, sfo.length)
        for j in range(sfo.nitems):
            pname = str_byt.to_str(buf[j0: j0+8])
            unit, dfmt, n_items = unpack('>3h', buf[j0+8:  j0+14])
            meta = type('', (), {})()
            meta.physunit = unit
            if unit in sfdics.unit_d.keys():
                meta.phys_unit = sfdics.unit_d[unit]
            meta.n_items = n_items
            meta.data_format = dfmt
            meta.status = unpack('>h', buf[j0+14:  j0+16])[0]
            logger.debug('PN: %s, j0: %d, unit: %d dfmt: %d n_items: %d, status: %d', pname, j0, unit, dfmt, n_items, meta.status)
            j0 += 16

            if dfmt in sfdics.fmt2len.keys(): # char variable
                dlen = sfdics.fmt2len[dfmt]
                bytlen = n_items * dlen
                meta.dmin = buf[j0  : j0+1]
                meta.dmax = buf[j0+1: j0+2]
                if len(meta.dmin) == 0:
                    meta.dmin = b' '
                if len(meta.dmax) == 0:
                    meta.dmax = b' '
                data = np.chararray((n_items,), itemsize=dlen, buffer=buf[j0+2: j0+2+bytlen])
                dj0 = 8 * ( (bytlen + 9)//8 )
                j0 += dj0
            elif dfmt in sfdics.fmt2typ.keys(): # number
                sfmt = sfdics.fmt2struc[dfmt]
                val_len = n_items + 2
                bytlen = val_len * np.dtype(sfmt).itemsize
                if dfmt == LOGICAL: # logical, bug if n_items > 1?
                    meta.dmin = unpack(sfmt, buf[j0  : j0+1])[0]
                    meta.dmax = unpack(sfmt, buf[j0+3: j0+4])[0]
                    data = unpack(sfmt, buf[j0+5: j0+6])[0]
                else:
                    data = np.ndarray((val_len, ), '>%s' %sfmt, buf[j0: j0+bytlen], order='F').copy()
                    meta.dmin = data[0]
                    meta.dmax = data[1]
                    data = np.squeeze(data[2:]) # no array if n_items=1
                dj0 = str_byt.next8(bytlen)
                j0 += dj0
            else: # faulty dfmt
                break
            par_d[pname] = sfobj.SFOBJ(data, sfho=meta)

            if j0 >= sfo.length:
                break
        return par_d


    def getlist(self, obj=None):
        """
        Returns a list of data-objects of a shotfile
        """
        if obj is None:
            obj = 'SIGNALS'
        else:
            obj = str_byt.to_str(obj)
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        dfmt = sfo.data_format
        if otyp != List:
            return None
        buf = self.getchunk(sfo.address, sfo.length)
        sfmt = sfdics.fmt2struc[dfmt]
        list_ids = unpack('>%d%s' %(sfo.nitems, sfmt), buf)

        return [self.getobjname(jid) for jid in list_ids]


    def getlist_by_type(self, obj_type=Signal):
        """
        Returns a list of names of all SF-objects of a given type (Signal, TimeBase)
        """
        return [lbl for lbl, sfo in self.sfh.items() if sfo.obj_type == obj_type]


    def getobjname(self, jobj):
        """
        Returns the object name for an inpur object ID
        """

        for lbl, sfo in self.sfh.items():
            if sfo.objid == jobj:
                return lbl


    def getdevice(self, lbl):
        """
        Returns a DEVICE object
        """

        return self.sfh[lbl].pars


    def get_ts06(self):
# Look anywhere in the diagsnostic's Devices
        ts06 = None
        for obj in self.parsets:
            if self.sfh[obj].obj_type == Device:
                ps = self.sfh[obj].pars
                if 'TS06' in ps.keys():
                    ts6 = ps['TS06']
                    if ts6 > 1e15:
                        ts06 = ts6
                        break
        return ts06


    def getts06(self, obj):
        """
        Reads the diagnostic internal TS06 from parameter set
        """
        ts06 = None
        if self.sfh[obj].obj_type == TimeBase:
            tb = obj
        else:
            for rel_obj in self.sfh[obj].relations:
                if self.sfh[rel_obj].obj_type == TimeBase: # related TB
                    tb = rel_obj
                    break
        if 'tb' in locals(): # No TB related
            obj2 = tb
        else: # try a direct relation
            obj2 = obj

        for rel_obj in self.sfh[obj2].relations:
            if self.sfh[rel_obj].obj_type == Device: # related device
                ps = self.sfh[rel_obj].pars
                if 'TS06' in ps.keys():
                    ts6 = ps['TS06']
                    if ts6 > 1e15:
                        ts06 = ts6
                        break

        logger.debug('getts06 %s %d', rel_obj, ts06)
        if ts06 is None:
            ts06 = self.get_ts06()
        if ts06 is None:
            ts06 = getcti_ts06(self.shot)
        return ts06


    def time_dim(self, obj):
        """
        Tells which SigGroup dimension is a time dimension
        """

        obj = str_byt.to_str(obj)
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        if otyp != SignalGroup:
            return None

        for rel in sfo.relations:
            if self.sfh[rel].obj_type == TimeBase:
                shape_arr = sfo.index[::-1][:sfo.num_dims]
                nt = self.sfh[rel].n_steps
                if shape_arr.count(nt) == 1:
                    return shape_arr.index(nt)
                else:
                    return None
        return None # If there is no timebase among the relations


    def time_first(self, obj):
        """
        Tells whether a SigGroup has time as first coordinate
        by comparing with the size of the related TBase
        """

        obj = str_byt.to_str(obj)
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        if otyp != SignalGroup:
            return False

        return (self.time_dim(obj) == 0)


    def time_last(self, obj):
        """
        Tells whether a SigGroup has time as first coordinate
        by comparing with the size of the related TBase
        """

        obj = str_byt.to_str(obj)
        sfo = self.sfh[obj]
        otyp = sfo.obj_type
        if otyp != SignalGroup:
            return False

        return (self.time_dim(obj) == sfo.num_dims-1)



"""Set value from kwargs, falling back to default.
Gives some flexibility for keynames.
"""
def parse_kw(lbl_in, kwargs, default=None):
    for key in lbl_in:
        if key in kwargs.keys():
            return kwargs[key]
    return default



#-------------------------------------------------------------------------------
# journal: Module to query the AUG journal
#
# History:
#
#    30.11.2016 jcf
#    15.10.2021 git   made py3 compatible, few auxiliary modules
#-----------------------------------------------------------------------------

import os, logging, datetime
import ctypes as ct
from aug_sfutils import str_byt, getlastshot, sfread

logger = logging.getLogger('aug_sfutils.journal')

# Select the correct library

libjouso = '/afs/ipp/aug-ads/diags/@sys/lib64/libjournal.so'
if not os.path.isfile(libjouso):
    libjouso = '/afs/ipp-garching.mpg.de/aug/ads-diags/amd64_sles15/lib64/libjournal.so'
libjournal   = ct.cdll.LoadLibrary(libjouso)

date_fmt = '%Y-%m-%d'


def getdict(struct):
    return dict((field, getattr(struct, field)) for field, _ in struct._fields_)


def anyshot():

    nshot = getlastshot.getlastshot()
    date_fmt = '%d%b%Y'
    today = datetime.datetime.now().strftime(date_fmt)
    date_last = 'No date'
    jou = sfread.SFREAD(nshot, 'JOU')
    if jou.status:
        date_last = jou.time.strftime(date_fmt)
    logger.info('Today is %s', today) 
    logger.info('Last shot %d on %s', nshot, date_last) 

    return(today == date_last)


class jouEntry(ct.Structure):
    """
    Class defining the ctypes structure for a full journal entry
    """
    _fields_ = [\
        ("shotno",    ct.c_int),
        ("program",   ct.c_char_p),
        ("clearance", ct.c_char_p),
        ("datum",     ct.c_char_p),
        ("time",      ct.c_char_p),
        ("leader",    ct.c_char_p),
        ("konf_il",   ct.c_char_p),
        ("konf_al",   ct.c_char_p),
        ("konf_snu",  ct.c_char_p),
        ("konf_sno",  ct.c_char_p),
        ("konf_dn",   ct.c_char_p),
        ("typ",       ct.c_char_p),
        ("useful",    ct.c_char_p),
        ("b_hmod",    ct.c_char_p),
        ("b_lmod",    ct.c_char_p),
        ("b_imprh",   ct.c_char_p),
        ("b_itb",     ct.c_char_p),
        ("b_blim",    ct.c_char_p),
        ("b_dlim",    ct.c_char_p),
        ("b_disb",    ct.c_char_p),
        ("b_disr",    ct.c_char_p),
        ("b_dise",    ct.c_char_p),
        ("b_vdeo",    ct.c_char_p),
        ("b_vdeu",    ct.c_char_p),
        ("b_nobd",    ct.c_char_p),
        ("b_res",     ct.c_char_p),
        ("b_fail",    ct.c_char_p),
        ("b_stable",  ct.c_char_p),
        ("b_sweep",   ct.c_char_p),
        ("b_cdh",     ct.c_char_p),
        ("b_run",     ct.c_char_p),
        ("vessel",    ct.c_char_p),
        ("coat",      ct.c_char_p),
        ("coatdate",  ct.c_char_p),
        ("proposal",  ct.c_char_p),
        ("yag",       ct.c_char_p),
        ("lagereg",   ct.c_char_p),
        ("lagecom1",  ct.c_char_p),
        ("lagecom2",  ct.c_char_p),
        ("formreg",   ct.c_char_p),
        ("formcom1",  ct.c_char_p),
        ("formcom2",  ct.c_char_p),
        ("remarks",   ct.c_char_p),
        ("flatb",     ct.c_double),
        ("flate",     ct.c_double),
        ("flatm",     ct.c_char_p),
        ("ip",        ct.c_double),
        ("ne",        ct.c_double),
        ("nediag",    ct.c_char_p),
        ("bt",        ct.c_double),
        ("q95",       ct.c_double),
        ("kappa",     ct.c_double),
        ("delrob",    ct.c_double),
        ("delrunt",   ct.c_double),
        ("disr",      ct.c_double),
        ("killergas", ct.c_char_p),
        ("impspez",   ct.c_char_p),
        ("imptime",   ct.c_double),
        ("neutr",     ct.c_double),
        ("kryop",     ct.c_char_p),
        ("glowtime",  ct.c_double),
        ("ioh",       ct.c_double),
        ("roh",       ct.c_double),
        ("gas_h",     ct.c_char_p),
        ("gas_d",     ct.c_char_p),
        ("gas_he",    ct.c_char_p),
        ("gas_ne",    ct.c_char_p),
        ("gas_ar",    ct.c_char_p),
        ("gas_n2",    ct.c_char_p),
        ("gas_kr",    ct.c_char_p),
        ("gas_xe",    ct.c_char_p),
        ("gas_cd4",   ct.c_char_p),
        ("gas_other", ct.c_char_p),
        ("valve_1",   ct.c_char_p),
        ("valve_2",   ct.c_char_p),
        ("valve_3",   ct.c_char_p),
        ("valve_4",   ct.c_char_p),
        ("valve_5",   ct.c_char_p),
        ("valve_6",   ct.c_char_p),
        ("valve_7",   ct.c_char_p),
        ("valve_8",   ct.c_char_p),
        ("valve_9",   ct.c_char_p),
        ("valve_10",  ct.c_char_p),
        ("valve_11",  ct.c_char_p),
        ("valve_12",  ct.c_char_p),
        ("valve_13",  ct.c_char_p),
        ("valve_14",  ct.c_char_p),
        ("valve_15",  ct.c_char_p),
        ("valve_16",  ct.c_char_p),
        ("valve_17",  ct.c_char_p),
        ("valve_18",  ct.c_char_p),
        ("valve_19",  ct.c_char_p),
        ("valve_20",  ct.c_char_p),
        ("valve_21",  ct.c_char_p),
        ("valve_22",  ct.c_char_p),
        ("erstfh",    ct.c_double),
        ("erstfb",    ct.c_double),
        ("erstfg",    ct.c_char_p),
        ("gasvalv",   ct.c_char_p),
        ("nbi1l",     ct.c_double),
        ("nbi1sp",    ct.c_double),
        ("nbi1g",     ct.c_char_p),
        ("nbi1b",     ct.c_double),
        ("nbi1e",     ct.c_double),
        ("nbi2l",     ct.c_double),
        ("nbi2sp",    ct.c_double),
        ("nbi2g",     ct.c_char_p),
        ("nbi2b",     ct.c_double),
        ("nbi2e",     ct.c_double),
        ("nbi3l",     ct.c_double),
        ("nbi3sp",    ct.c_double),
        ("nbi3g",     ct.c_char_p),
        ("nbi3b",     ct.c_double),
        ("nbi3e",     ct.c_double),
        ("nbi4l",     ct.c_double),
        ("nbi4sp",    ct.c_double),
        ("nbi4g",     ct.c_char_p),
        ("nbi4b",     ct.c_double),
        ("nbi4e",     ct.c_double),
        ("nbi5l",     ct.c_double),
        ("nbi5sp",    ct.c_double),
        ("nbi5g",     ct.c_char_p),
        ("nbi5b",     ct.c_double),
        ("nbi5e",     ct.c_double),
        ("nbi6l",     ct.c_double),
        ("nbi6sp",    ct.c_double),
        ("nbi6g",     ct.c_char_p),
        ("nbi6b",     ct.c_double),
        ("nbi6e",     ct.c_double),
        ("nbi7l",     ct.c_double),
        ("nbi7sp",    ct.c_double),
        ("nbi7g",     ct.c_char_p),
        ("nbi7b",     ct.c_double),
        ("nbi7e",     ct.c_double),
        ("nbi8l",     ct.c_double),
        ("nbi8sp",    ct.c_double),
        ("nbi8g",     ct.c_char_p),
        ("nbi8b",     ct.c_double),
        ("nbi8e",     ct.c_double),
        ("ecrh1l",    ct.c_double),
        ("ecrh1f",    ct.c_double),
        ("e1m",       ct.c_char_p),
        ("ecrh1b",    ct.c_double),
        ("ecrh1e",    ct.c_double),
        ("ecrh2l",    ct.c_double),
        ("ecrh2f",    ct.c_double),
        ("e2m",       ct.c_char_p),
        ("ecrh2b",    ct.c_double),
        ("ecrh2e",    ct.c_double),
        ("ecrh3l",    ct.c_double),
        ("ecrh3f",    ct.c_double),
        ("e3m",       ct.c_char_p),
        ("ecrh3b",    ct.c_double),
        ("ecrh3e",    ct.c_double),
        ("ecrh4l",    ct.c_double),
        ("ecrh4f",    ct.c_double),
        ("e4m",       ct.c_char_p),
        ("ecrh4b",    ct.c_double),
        ("ecrh4e",    ct.c_double),
        ("ecrh5l",    ct.c_double),
        ("ecrh5f",    ct.c_double),
        ("e5m",       ct.c_char_p),
        ("ecrh5b",    ct.c_double),
        ("ecrh5e",    ct.c_double),
        ("ecrh6l",    ct.c_double),
        ("ecrh6f",    ct.c_double),
        ("e6m",       ct.c_char_p),
        ("ecrh6b",    ct.c_double),
        ("ecrh6e",    ct.c_double),
        ("ecrh7l",    ct.c_double),
        ("ecrh7f",    ct.c_double),
        ("e7m",       ct.c_char_p),
        ("ecrh7b",    ct.c_double),
        ("ecrh7e",    ct.c_double),
        ("ecrh8l",    ct.c_double),
        ("ecrh8f",    ct.c_double),
        ("e8m",       ct.c_char_p),
        ("ecrh8b",    ct.c_double),
        ("ecrh8e",    ct.c_double),
        ("icrh1l",    ct.c_double),
        ("icrh1f",    ct.c_double),
        ("i1m",       ct.c_char_p),
        ("icrh1b",    ct.c_double),
        ("icrh1e",    ct.c_double),
        ("icrh2l",    ct.c_double),
        ("icrh2f",    ct.c_double),
        ("i2m",       ct.c_char_p),
        ("icrh2b",    ct.c_double),
        ("icrh2e",    ct.c_double),
        ("icrh3l",    ct.c_double),
        ("icrh3f",    ct.c_double),
        ("i3m",       ct.c_char_p),
        ("icrh3b",    ct.c_double),
        ("icrh3e",    ct.c_double),
        ("icrh4l",    ct.c_double),
        ("icrh4f",    ct.c_double),
        ("i4m",       ct.c_char_p),
        ("icrh4b",    ct.c_double),
        ("icrh4e",    ct.c_double),
        ("nbi4m",     ct.c_double),
        ("icrh4m",    ct.c_double),
        ("ecrh4m",    ct.c_double),
        ("nbb1b",     ct.c_double),
        ("nbb1e",     ct.c_double),
        ("pheattot",  ct.c_double),
        ("status",    ct.c_char_p),
        ("notice",    ct.c_char_p),
        ("upddate",   ct.c_char_p),
        ("cryoreg",   ct.c_char_p),
        ("programdir",ct.c_char_p),
        ("crossval",  ct.c_char_p),
        ("crosstyp",  ct.c_char_p),
        ("recipe",    ct.c_char_p),
        ("saddlecr",  ct.c_int),
        ("saddleip",  ct.c_double),
        ("saddleci",  ct.c_double),
           ]


def journalSearch(SQLquery, onlyUseful=1, shotRange= [0, 0]):
    """Search entries in the logbook

    Syntax:

    sf.journalSearch(SQLquery, [Options])

    Arguments and Options:
       SQLquery               string    Query in SQL syntax (required)
       onlyUseful=0|1         int       Flag, if only useful plasmashots ar te be searched (default: yes)
       shotRange=[low, high]  [int,int] Only search within this shot range

    Returns:
           array of dictionary of shot entries (or None)
    """

# Check Arguments

    assert isinstance(SQLquery, str),  'SQLquery: string expected'

    if onlyUseful is not None:
        assert isinstance(onlyUseful, int), 'onlyUseful: int expected'
        libjournal.jouSetOnlyUseful(ct.c_int(onlyUseful))

    if shotRange is not None:
        libjournal.jouSetShotRange(shotRange[0], shotRange[1])

# Search

    nResults = ct.c_int(0)
    rc = libjournal.jouSearch(SQLquery, ct.byref(nResults))
    if rc or nResults == 0:
        return

    entries = (jouEntry * nResults.value)()
    rc = libjournal.jouFetchEntries(ct.byref(entries))

    pyEntries = [getdict(entries[i]) for i in range(nResults.value)]
    return pyEntries


def setShotRange(shot1, shot2):
    """Set the shot range for the following searches

    Syntax:

    sf.setShotRange(shot1, shot2)

    Arguments:
    shot1    int  lower bound (0 means unset)
    shot2    int  upper bound (0 means unset)

    Returns:
    nothing
    """

# Check arguments

    assert isinstance(shot1, int), 'shot1: int expected'
    assert isinstance(shot2, int), 'shot1: int expected'

# Set

    libjournal.jouSetShotRange(shot1, shot2)


def setOnlyUseful(flag):
    """Set the flag, if only useful plasma discharges should be searched for

    Syntax:

    sf.setOnlyUseful(flag)

    Arguments:
    flag    int  (0 or 1)

    Returns:
    nothing
    """

# Check arguments

    assert isinstance(flag, int), 'flag: int expected'

# Set

    libjournal.jouSetOnlyUseful(flag)


def getNResults():
    """Get the number of results from the previous search

    Syntax:

    sf.getNResults

    Returns:
    number of results
    """

    n = libjournal.jouGetNResults()
    return n


def getEntryForShot(shot):
    """Get the parameters for one shot

    Syntax:

    sf.getEntryForShot(shot)

    Returns:
    Directory with parameters
    """

# Check arguments

    assert isinstance(shot, int), 'shot: int expected'

# Get 

    entry = (jouEntry)()
    rc = libjournal.jouGetEntryForShot(shot, ct.byref(entry))

    return getdict(entry)


def searchSession(date=None, fmt=None, onlyUseful=1):
    """Get parameters for one session

    Syntax:

    sf.searchSession( date='', [OPTIONS] )

    Arguments and Options:

    date            str   Date of the session
    fmt=format      str   Format of the date string (default: YYYY-MM-DD)
    onlyUseful=0|1  int   Flag, if only useful plasmashots ar te be searched (default: yes)

    Returns:
        array of dictionary of shot entries (or None)
    """

# Check arguments

    date, fmt = checkDateFmt(date, fmt)

    if onlyUseful is not None:
        assert isinstance(onlyUseful, int), 'onlyUseful: int expected'
        libjournal.jouSetOnlyUseful(ct.c_int(onlyUseful))

# Get 

    rc = libjournal.jouSearchSession(date, fmt)
    if rc:
        return None

    nResults = libjournal.jouGetNResults()
    if nResults == 0:
        return None

    entries = (jouEntry * nResults)();
    rc = libjournal.jouFetchEntries(ct.byref(entries));

    pyEntries = [getdict(entries[i]) for i in range(nResults)]
    return pyEntries


def getSession(date=None, fmt=None, next=True):
    """Get next or previous session by date.

    Syntax:
        sf.getSession(date='', fmt='', next=True)

    Input:
        date=date  str  Date (default: today)
        fmt=fmt    str  Format of the date string (default: 'YYYY-MM-DD')
        next=True  bool If True, takes following date, else previous
    Output:
        Date (string) of next/previous session
    """

    date, fmt = checkDateFmt(date, fmt)
    if date is None:
        return None

# Get 

    closeDate = date
    if next:
        rc = libjournal.jouGetNextSession(date, closeDate, fmt)
    else:
        rc = libjournal.jouGetPrevSession(date, closeDate, fmt)
    if rc:
        return None

    return closeDate[:-1]


def getPrevSession(*args, **kwargs):
    """Get the date of the session preceding date

    Syntax:
        sf.getPreviousSession(date='', fmt='')

    Input:
        date=date  str  Date (default: today)
        fmt=fmt    str  Format of the date string (default: 'YYYY-MM-DD')

    Output:
        Date (string) of previous session
    """

    kwargs['next'] = False
    return getSession(*args, **kwargs)


def getNextSession(*args, **kwargs):
    """Get the date of the session preceding date

    Syntax:
        sf.getNextSession(date='', fmt='')

    Input:
        date=date  str  Date (default: today)
        fmt=fmt    str  Format of the date string (default: 'YYYY-MM-DD')

    Output:
        Date (string) of following session
    """

    return getSession(*args, **kwargs)


def checkDateFmt(date=None, fmt=None):
    """Checks the format of the input date.
    Default date is today.
    """
    if date is None:
        today = datetime.date.today()
        date = today.strftime(date_fmt)

    try:
        date = str_byt.to_byt(date)
    except:
        logger.error('date must be strong or byte array')
        return None, None

    if fmt is None:
        fmt = b''
    else:
        try:
            fmt = str_byt.to_byt(fmt)
        except:
            logger.error('fmt must be string or byte array')
            return None, None

    date = ct.create_string_buffer(date)
    fmt  = ct.create_string_buffer(fmt)

    return date, fmt


def isExpDay(date=None, fmt=None):
    """
    Syntax:
        sf.isExpDay(date=..., fmt=...)

    Input:
        date (optional): str   Date in format fmt. If None, today is assumed
        fmt  (optinal): str    Date format
    Output:
        True if day=date is/was an experimental day, False otherwise
    """

    session = searchSession(date=date, fmt=fmt, onlyUseful=0) # Today
    return (session is not None)


def getLastShot(onlyUseful=0):
    """
    Syntax:
        sf.getLastShot(onlyUseful=...)

    Input:
        onlyUseful (optional) | int | Values: 0 (all shots, default), 1 (useful shots)
    Output:
        Last AUG shot  | int
    """

    today = datetime.date.today()
    today_sf = today.strftime(date_fmt)
    session = searchSession(onlyUseful=onlyUseful) # Today
    if session is None:
        date_last = getPrevSession()
        session = searchSession(date=date_last, onlyUseful=onlyUseful)
    return(session[-1]['shotno'])



"""
Equilibrium utilities library
"""

import sys, time, logging, datetime
import numpy as np
from scipy.ndimage.interpolation import map_coordinates
from scipy.interpolate import UnivariateSpline, interp1d, InterpolatedUnivariateSpline, RectBivariateSpline

logger = logging.getLogger('aug_sfutils.mapeq')
#logger.setLevel(logging.INFO)

def compare_version(a, b):
    from pkg_resources import parse_version
    return parse_version(a) > parse_version(b)

def get_nearest_index(tim_eq, tarr):
    """Find nearest time index for a given time.
    For internal use only.
    """

    tim_eq = np.atleast_1d(np.array(tim_eq, dtype=np.float32))
    if len(tim_eq) == 1:
        unique_idx = [0]
        index_idx = {0: [0]}
    else:
        tarr = np.clip(tarr, tim_eq[0], tim_eq[-1])
        idx = interp1d(tim_eq, np.arange(len(tim_eq)), kind='nearest', assume_sorted=True)(tarr)
        idx = idx.astype(np.int32)
        unique_idx = np.unique(idx)
        index_idx = {}
        for i in unique_idx:
            (index_idx[i], ) = np.where(idx == i)

    return unique_idx, index_idx


def rho2rho(eqm, rho_in, t_in=None, \
           coord_in='rho_pol', coord_out='rho_tor', extrapolate=False):

    """Mapping from/to rho_pol, rho_tor, r_V, rho_V, Psi, r_a
    r_V is the STRAHL-like radial coordinate

    Input
    ----------
    eqm: equilibrium object
    t_in : float or 1darray or None
        time (eqm.time if None)
    rho_in : float, ndarray
        radial coordinates, 1D (time constant) or 2D+ (time variable) of size (nt,nx,...)
    coord_in:  str ['rho_pol', 'rho_tor' ,'rho_V', 'r_V', 'Psi','r_a']
        input coordinate label
    coord_out: str ['rho_pol', 'rho_tor' ,'rho_V', 'r_V', 'Psi','r_a']
        output coordinate label
    extrapolate: bool
        extrapolate rho_tor, r_V outside the separatrix

    Output
    -------
    rho : 2d+ array (nt, nr, ...)
    converted radial coordinate
    """

    logger.debug('Start rho2rho')
    if t_in is None:
        t_in = eqm.time

    tarr = np.atleast_1d(t_in)
    rho = np.atleast_1d(rho_in)

    nt_in = np.size(tarr)

    if rho.ndim == 1:
        rho = np.tile(rho, (nt_in, 1))

# Trivial case
    if coord_out == coord_in: 
        return rho
 
    unique_idx, index_idx =  get_nearest_index(eqm.time, tarr)

    if coord_in in ['rho_pol', 'Psi']:
        label_in = eqm.pfl
        lbl_in = 'pfl'
    elif coord_in == 'rho_tor':
        label_in = eqm.tfl
        lbl_in = 'tfl'
    elif coord_in in ['rho_V','r_V']:
        label_in = eqm.vol
        R0 = eqm.Rmag
        lbl_in = 'vol'
    elif coord_in == 'r_a':
        R, _ = rhoTheta2rz(eqm, eqm.pfl, [0, np.pi], coord_in='Psi')
        label_in = (R[:, 0] - R[:, 1]).T**2
        lbl_in = 'r_a'
    else:
        raise Exception('unsupported input coordinate')

    if coord_out in ['rho_pol', 'Psi']:
        label_out = eqm.pfl
        lbl_out = 'pfl'
    elif coord_out == 'rho_tor':
        label_out = eqm.tfl
        lbl_out = 'tfl'
    elif coord_out in ['rho_V','r_V']:
        label_out = eqm.vol
        R0 = eqm.Rmag
        lbl_out = 'vol'
    elif coord_out == 'r_a':
        R, _ = rhoTheta2rz(eqm, eqm.pfl[unique_idx], [0, np.pi], \
                    t_in=eqm.time[unique_idx], coord_in='Psi')
        label_out = np.zeros_like(eqm.pfl)
        label_out[unique_idx, :] = (R[:, 0] - R[:, 1])**2
        lbl_out = 'r_a'
    else:
        raise Exception('unsupported output coordinate')

    psi_sign = np.sign(eqm.psi_fac)
    PFL  = psi_sign*eqm.pfl
    PSIX = psi_sign*eqm.psix
    PSI0 = psi_sign*eqm.psi0
    rho_output = np.zeros_like(rho)

    for i in unique_idx:
  
# Calculate a normalized input and output flux 

        mag_out, sep_out = np.interp([PSI0[i], PSIX[i]],  PFL[i], label_out[i])
        if lbl_in != lbl_out:
            mag_in, sep_in = np.interp([PSI0[i], PSIX[i]],  PFL[i], label_in[i])
        else:
            mag_in = mag_out
            sep_in = sep_out

        if (abs(sep_out - mag_out) < 1e-4) or (abs(sep_in - mag_in) < 1e-4): #corrupted timepoint
            continue

# Normalize between 0 and 1
        rho_out = (label_out[i] - mag_out)/(sep_out - mag_out)
        rho_in  = (label_in [i] - mag_in )/(sep_in  - mag_in )

        rho_out[(rho_out > 1) | (rho_out < 0)] = 0  #remove rounding errors
        rho_in[ (rho_in  > 1) | (rho_in  < 0)] = 0

        rho_out = np.r_[np.sqrt(rho_out), 1]
        rho_in  = np.r_[np.sqrt(rho_in ), 1]

        ind = (rho_out==0) | (rho_in==0)
        rho_out, rho_in = rho_out[~ind], rho_in[~ind]
        
# Profiles can be noisy!  smooth spline must be used
        sortind = np.unique(rho_in, return_index=True)[1]
        w = np.ones_like(sortind)*rho_in[sortind]
        w = np.r_[w[1]/2, w[1:], 1e3]
        ratio = rho_out[sortind]/rho_in[sortind]
        rho_in = np.r_[0, rho_in[sortind]]
        ratio = np.r_[ratio[0], ratio]

        s = UnivariateSpline(rho_in, ratio, w=w, k=4, s=5e-3,ext=3)  #BUG s = 5e-3 can be sometimes too much, sometimes not enought :( 

        jt = index_idx[i]
        rho_ = np.copy(rho[jt])

        r0_in = 1
        if coord_in == 'r_V' :
            r0_in  = np.sqrt(sep_in/ (2*np.pi**2*R0[i]))
        r0_out = 1
        if coord_out == 'r_V' :
            r0_out = np.sqrt(sep_out/(2*np.pi**2*R0[i]))

        if coord_in == 'Psi' :
            rho_  = np.sqrt(np.maximum(0, (rho_ - eqm.psi0[i])/(eqm.psix[i] - eqm.psi0[i])))

# Evaluate spline

        rho_output[jt] = s(rho_.flatten()/r0_in).reshape(rho_.shape)*rho_*r0_out/r0_in

        if np.any(np.isnan(rho_output[jt])):  # UnivariateSpline failed
            rho_output[jt] = np.interp(rho_/r0_in, rho_in, ratio)*rho_*r0_out/r0_in
            
        if not extrapolate:
            rho_output[jt] = np.minimum(rho_output[jt],r0_out) # rounding errors

        rho_output[jt] = np.maximum(0,rho_output[jt]) # rounding errors

        if coord_out  == 'Psi':
            rho_output[jt]  = rho_output[jt]**2*(eqm.psix[i] - eqm.psi0[i]) + eqm.psi0[i]

    logger.debug('End rho2rho')
    return rho_output


def rz2brzt(eqm, r_in, z_in, t_in=None):

    """calculates Br, Bz, Bt profiles
    Input
    ----------
    eqm: equilibrium object
    r_in : ndarray
        R coordinates 
        1D, size(nrz_in) or 2D, size (nt, nrz_in)
    z_in : ndarray
        Z coordinates 
        1D, size(nrz_in) or 2D, size (nt, nrz_in)
    t_in : float or 1darray or None
        time (eqm.time if None)

    Output
    -------
    interpBr : ndarray
        profile of Br(nt, nrz_in)
    interpBz : ndarray
        profile of Bz(nt, nrz_in)
    interpBt : ndarray
        profile of Bt(nt, nrz_in)
    """

    if t_in is None:
        t_in = eqm.time
        
    tarr = np.atleast_1d(t_in)
    r_in = np.atleast_1d(r_in)
    z_in = np.atleast_1d(z_in)

    nt = np.size(tarr)
    nrz_in = r_in.shape[-1]

    if r_in.shape != z_in.shape:
        logger.error('r_in and z_in must have the same shape')
        return

    if r_in.ndim == 2 and r_in.shape[0] != nt:
        logger.error('Time array missmatching first dim of r_in')
        return

    if np.size(r_in, 0)!= nt:
        r_in = np.tile(r_in, nt).reshape((nt, nrz_in))
        z_in = np.tile(z_in, nt).reshape((nt, nrz_in))

    interpBr = np.zeros((nt, nrz_in), dtype=np.float32)
    interpBz = np.zeros((nt, nrz_in), dtype=np.float32)
    interpBt = np.zeros((nt, nrz_in), dtype=np.float32)

    from scipy.constants import mu_0
    nr, nz = len(eqm.Rmesh), len(eqm.Zmesh)
    dr = eqm.Rmesh[1] - eqm.Rmesh[0]
    dz = eqm.Zmesh[1] - eqm.Zmesh[0]

    unique_idx, index_idx =  get_nearest_index(eqm.time, tarr)

    for i in unique_idx:

        jt = index_idx[i]
        Psi = eqm.pfm[:, :, i]
        Bpol = np.gradient(Psi, dr, dz)/eqm.Rmesh[:, None]
        fBt = interp1d(eqm.pfl[i], eqm.jpol[i], kind='linear', fill_value='extrapolate')
        Bt = fBt(Psi)*mu_0/eqm.Rmesh[:, None]
        r = r_in[jt] #np.squeeze(r_in[jt])
        z = z_in[jt] #np.squeeze(z_in[jt])
        f_br = RectBivariateSpline(eqm.Rmesh, eqm.Zmesh,  Bpol[1], kx=2, ky=2)
        f_bz = RectBivariateSpline(eqm.Rmesh, eqm.Zmesh, -Bpol[0], kx=2, ky=2)
        f_bt = RectBivariateSpline(eqm.Rmesh, eqm.Zmesh,       Bt, kx=2, ky=2)
        interpBr[jt] = f_br(r, z, grid=False)
        interpBz[jt] = f_bz(r, z, grid=False)
        interpBt[jt] = f_bt(r, z, grid=False)

    return -interpBr/abs(eqm.psi_fac), \
           -interpBz/abs(eqm.psi_fac), \
           interpBt/(2.*np.pi) * eqm.psi_sign * eqm.phi_sign


def Bmesh(eqm, t_in=None):

    """calculates Br, Bz, Bt profiles
    Input
    ----------
    eqm: equilibrium object
    t_in : float or 1darray or None
        time (eqm.time if None)

    Output
    -------
    Br : ndarray
        profile of Br on the PFM grid
    Bz : ndarray
        profile of Bz on the PFM grid
    interpBt : ndarray
        profile of Bt on the PFM grid
    """

    if t_in is None:
        t_in = eqm.time

    tarr = np.atleast_1d(t_in)
    nt = len(tarr)

# Poloidal current 

    from scipy.constants import mu_0

    nr = len(eqm.Rmesh)
    nz = len(eqm.Zmesh)
    dr = eqm.Rmesh[1] - eqm.Rmesh[0]
    dz = eqm.Zmesh[1] - eqm.Zmesh[0]

    unique_idx, index_idx =  get_nearest_index(eqm.time, tarr)

    Br = np.zeros((nt, nr, nz), dtype=np.float32)
    Bz = np.zeros_like(Br)
    Bt = np.zeros_like(Br)
    for i in unique_idx:

        jt = index_idx[i]
        Psi = eqm.pfm[:, :, i]
# Eq 12 in Coco paper
        Bpol = np.gradient(Psi, dr, dz)/eqm.Rmesh[:, None]
        Br[jt] =  Bpol[1]
        Bz[jt] = -Bpol[0]
        fBt = interp1d(eqm.pfl[i], eqm.jpol[i], kind='linear', fill_value='extrapolate')
        Bt[jt] = fBt(Psi)*mu_0/eqm.Rmesh[:, None]

    return -Br/abs(eqm.psi_fac), \
           -Bz/abs(eqm.psi_fac), \
           Bt/(2.*np.pi) * eqm.psi_sign * eqm.phi_sign


def rz2rho(eqm, r_in, z_in, t_in=None, coord_out='rho_pol', extrapolate=True):

    """
    Equilibrium mapping routine, map from R,Z -> rho (pol,tor,r_V,...)
    Fast for a large number of points

    Input
    ----------
    eqm: equilibrium object
    t_in : float or 1darray or None
        time (eqm.time if None)
    r_in : ndarray
        R coordinates 
        1D (time constant) or 2D+ (time variable) of size (nt,nx,...)
    z_in : ndarray
        Z coordinates 
        1D (time constant) or 2D+ (time variable) of size (nt,nx,...)
    coord_out: str
        mapped coordinates - rho_pol,  rho_tor, r_V, rho_V, Psi
    extrapolate: bool
        extrapolate coordinates (like rho_tor) for values larger than 1

    Output
    -------
    rho : 2D+ array (nt,nx,...)
    Magnetics flux coordinates of the points
    """

    if t_in is None:
        t_in = eqm.time

    tarr = np.atleast_1d(t_in)
    r_in = np.atleast_2d(r_in)
    z_in = np.atleast_2d(z_in)

    dr = (eqm.Rmesh[-1] - eqm.Rmesh[0])/(len(eqm.Rmesh) - 1)
    dz = (eqm.Zmesh[-1] - eqm.Zmesh[0])/(len(eqm.Zmesh) - 1)

    nt_in = np.size(tarr)

    if np.size(r_in, 0) == 1:
        r_in = np.tile(r_in, (nt_in, 1))
    if np.size(z_in, 0) == 1:
        z_in = np.tile(z_in, (nt_in, 1))

    if r_in.shape!= z_in.shape:
        raise Exception('Wrong shape of r_in or z_in')
    
    if np.size(r_in,0) != nt_in:
        raise Exception('Wrong shape of r_in %s, nt=%d' %(str(r_in.shape), nt_in))
    if np.size(z_in,0) != nt_in:
        raise Exception('Wrong shape of z_in %s, nt=%d' %(str(z_in.shape), nt_in))
    if np.shape(r_in) != np.shape(z_in):
        raise Exception( 'Not equal shape of z_in and r_in %s,%s'\
                        %(str(z_in.shape), str(z_in.shape)) )

    Psi = np.empty((nt_in,)+r_in.shape[1:], dtype=np.single)

    scaling = np.array([dr, dz])
    offset  = np.array([eqm.Rmesh[0], eqm.Zmesh[0]])
    
    unique_idx, index_idx =  get_nearest_index(eqm.time, tarr)

    for i in unique_idx:
        jt = index_idx[i]
        coords = np.array((r_in[jt], z_in[jt]))
        index = ((coords.T - offset) / scaling).T
        Psi[jt] =  map_coordinates(eqm.pfm[:, :, i], index, mode='nearest',
                                   order=2, prefilter=True)

    rho_out = rho2rho(eqm, Psi, t_in=t_in, extrapolate=extrapolate, \
              coord_in='Psi', coord_out=coord_out)

    return rho_out


def rho2rz(eqm, rho_in, t_in=None, coord_in='rho_pol', all_lines=False):

    """
    Get R, Z coordinates of a flux surfaces contours

    Input
    ----------
    eqm: equilibrium object
    t_in : float or 1darray
        time (eqm.time if None)
    rho_in : 1darray,float
        rho coordinates of the searched flux surfaces
    coord_in: str
        mapped coordinates - rho_pol or rho_tor
    all_lines: bool:
        True - return all countours , False - return longest contour

    Output
    -------
    rho : array of lists of arrays [npoinst,2]
        list of times containg list of surfaces for different rho 
        and every surface is decribed by 2d array [R,Z]
    """

    if t_in is None:
        t_in = eqm.time

    tarr  = np.atleast_1d(t_in)
    rhoin = np.atleast_1d(np.array(rho_in, dtype=np.float32))

    rho_in = rho2rho(eqm, rhoin, t_in=t_in, \
             coord_in=coord_in, coord_out='Psi', extrapolate=True )

    from matplotlib import _contour
    import matplotlib

    nt = len(tarr)
    nr = len(eqm.Rmesh)
    nz = len(eqm.Zmesh)

    R, Z = np.meshgrid(eqm.Rmesh, eqm.Zmesh)
    Rsurf = np.empty(nt, dtype='object')
    zsurf = np.empty(nt, dtype='object')
    unique_idx, index_idx = get_nearest_index(eqm.time, tarr)

    for i in unique_idx:

        jt = index_idx[i]
        Flux = rho_in[jt[0]]

# matplotlib's contour creation
        try:
            cgen = _contour.QuadContourGenerator(R, Z, eqm.pfm[:nr, :nz, i].T, None, True, 0)
        except Exception as e:
            logger.exception(R.shape, Z.shape, eqm.pfm[:nr, :nz, i].T.shape)
            raise e 

        Rs_t = []
        zs_t = []

        for jfl, fl in enumerate(Flux):
            verts = cgen.create_contour(fl)
            if compare_version(matplotlib.__version__, '3.4.3'):
                """After MPL ver 3.4.3 the return type is vertices and codes as a list. Prior it was just vertices"""
                segs = verts[0]
            else:
                segs = verts
            n_segs = len(segs)
            if n_segs == 0:
                if fl == eqm.psi0[i]:
                    Rs_t.append(np.atleast_1d(eqm.Rmag[i]))
                    zs_t.append(np.atleast_1d(eqm.Zmag[i]))
                else:
                    Rs_t.append(np.zeros(1))
                    zs_t.append(np.zeros(1))
                continue
            elif all_lines: # for open field lines
                line = np.vstack(segs)
            else:  #longest filed line, default
                line = []
                for l in segs:
                    if len(l) > len(line):
                        line = l

            R_surf, z_surf = list(zip(*line))
            R_surf = np.array(R_surf, dtype = np.float32)
            z_surf = np.array(z_surf, dtype = np.float32)
            if not all_lines:
                ind = (z_surf >= eqm.Zunt[i])
                if len(ind) > 1:
                    R_surf = R_surf[ind]
                    z_surf = z_surf[ind]
            Rs_t.append(R_surf)
            zs_t.append(z_surf)
   
        for j in jt:
            Rsurf[j] = Rs_t
            zsurf[j] = zs_t

    return Rsurf, zsurf


def cross_surf(eqm, rho=1., r_in=1.65, z_in=0, line_m=2, theta_in=0, t_in=None, coord_in='rho_pol'):

    """
    Computes intersections of a line with any flux surface.

    Input:
    ----------
    eqm: equilibrium object
    rho: float or 1D array(nt)
        coordinate of the desired flux surface
    t_in: float or 1darray(nt)
        time (eqm.time in case of None)
    r_in: float
        R position of the point
    z_in: float
        z position of the point
    theta_in: float
        angle of the straight line with respect to horizontal-outward
    coord_in:  str ['rho_pol', 'rho_tor' ,'rho_V', 'r_V', 'Psi','r_a']
        input coordinate label

    Output:
    ----------
    Rout: 3darray size(nt, 2)
        R-position of intersections
    zout: 3darray size(nt, 2)
        z-position of intersections
    """

    if t_in is None:
        t_in = eqm.time

    tarr = np.atleast_1d(t_in)
    rho  = np.atleast_1d(rho)
    r_in = np.float32(r_in)
    z_in = np.float32(z_in)

    unique_idx, index_idx = get_nearest_index(eqm.time, tarr)

    n_line = int(200*line_m) + 1 # 1 cm, but then there's interpolation!
    t = np.linspace(-line_m, line_m, n_line, dtype=np.float32)

    line_r = t*np.cos(theta_in) + r_in
    line_z = t*np.sin(theta_in) + z_in

    rho_line = rz2rho(eqm, line_r, line_z, t_in=eqm.time[unique_idx], \
                           coord_out=coord_in, extrapolate=True)

    nt_in = len(tarr)
    Rout = np.nan*np.ones((nt_in, 2), dtype=np.float32)
    zout = np.nan*np.ones((nt_in, 2), dtype=np.float32)
    for i, ii in enumerate(unique_idx):
        jt = index_idx[ii]
        if len(rho) == 1:
            rho_ref = rho[0]
        else:
            rho_ref = rho[jt[0]]
        ind_cross = np.argwhere(np.diff(np.sign(rho_line[i] - rho_ref))).flatten()

        if len(ind_cross) > 2:
            logger.warning('More than 2 intersections at t=%.3f', tarr[jt[0]])
            continue

        if len(ind_cross) == 0:
            logger.warning('No intersections found at t=%.3f', tarr[jt[0]])
            continue

        pos = 0
        for j in ind_cross:
            if rho_line[i, j + 1] > rho_line[i, j]:
                ind = [j, j + 1]
            else:
                ind = [j + 1, j]
            ztmp = np.interp(rho_ref, rho_line[i, ind], line_z[ind])
            if ztmp >= eqm.Zunt[ii] and ztmp <= eqm.Zoben[ii]:
                zout[jt, pos] = ztmp
                Rout[jt, pos] = np.interp(rho_ref, rho_line[i, ind], line_r[ind])
                pos += 1
    del line_r, line_z, rho_line

    return Rout, zout


def cross_sep(eqm, r_in=1.65, z_in=0, line_m=2., theta_in=0, t_in=None):

    """
    Computes intersections of a line with any flux surface.

    Input:
    ----------
    eqm: equilibrium object
    t_in: float or 1darray
        time (eqm.time in case of None)
    r_in: float
        R position of the point
    z_in: float
        z position of the point
    theta_in: float
        angle of the straight line with respect to horizontal-outward

    Output:
    ----------
    Rout: 3darray size(nt, nx, 2)
        R-position of intersections
    zout: 3darray size(nt, nx, 2)
        z-position of intersections
    """

    if t_in is None:
        t_in = eqm.time
    tarr = np.atleast_1d(t_in)
    unique_idx, index_idx = get_nearest_index(eqm.time, tarr)

    psi_sep = eqm.pfl[unique_idx, max(eqm.lpfp)]

    return cross_surf(eqm, rho=psi_sep, r_in=r_in, z_in=z_in, line_m=line_m, theta_in=theta_in, t_in=t_in, coord_in='Psi')


def rhoTheta2rz(eqm, rho, theta_in, t_in=None, coord_in='rho_pol', n_line=201):
    
    """
    This routine calculates the coordinates R, z of the intersections of
    a ray starting at the magnetic axis with fluxsurfaces 
    at given values of some radial coordinate.
    (slower than countours)

    Input:
    ----------
    rho: float or 1D array (nr) or nD array (nt,nr,...)
        coordinate of the desired flux surface inside LCFS!
    t_in: float or 1darray
        time (eqm.time in case of None)
    theta_in: float or 1D array n_theta
        angle of the straight line with respect to horizontal-outward, in radians!!
    coord_in:  str ['rho_pol', 'rho_tor' ,'rho_V', 'r_V', 'Psi','r_a']
        input coordinate label

    Output:
    ----------
    R ,  z: 3d+ array size(nt, n_theta, nr,...)
    """

    if t_in is None:
        t_in = eqm.time

    tarr = np.atleast_1d(t_in)

    nt_in = len(tarr)

    rho  = np.atleast_1d(rho)
    if rho.ndim == 1:
        rho = np.tile(rho, (nt_in, 1))

    theta_in = np.atleast_1d(theta_in)[:, None]
    ntheta = len(theta_in)

    unique_idx, index_idx = get_nearest_index(eqm.time, tarr)

# n_line = 201 <=> 5 mm, but then there's interpolation!

    line_r = np.empty((len(unique_idx), ntheta, n_line))
    line_z = np.empty((len(unique_idx), ntheta, n_line))

    line_m = .9 # line length: 0.9 m
    t = np.linspace(0, 1, n_line)**.5*line_m
    c, s = np.cos(theta_in), np.sin(theta_in)
   
    tmpc = c*t
    tmps = s*t
    for i, ii in enumerate(unique_idx):
        line_r[i] = tmpc + eqm.Rmag[ii]
        line_z[i] = tmps + eqm.Zmag[ii]
 
    rho_line = rz2rho(eqm, line_r, line_z, eqm.time[unique_idx], \
                           coord_out=coord_in , extrapolate=True)

    R = np.empty((nt_in, ntheta) + rho.shape[1:], dtype=np.float32)
    z = np.empty((nt_in, ntheta) + rho.shape[1:], dtype=np.float32)

    jcoco = eqm.cocos%10 - 1
    if coord_in == 'Psi':
        rho_line[:,:,0] = eqm.psi0[unique_idx][:,None]
        rho_line *= -eqm.ip_ccw*sigma['bp'][jcoco]
        rho      *= -eqm.ip_ccw*sigma['bp'][jcoco]
    else:
        #solve some issues very close to the core
        rho_line[:,:,0] = 0

    for i, ii in enumerate(unique_idx):
        jt = index_idx[ii]
        for k in range(ntheta):
            rho_lin = rho_line[i, k]
            (tmp, ) = np.where(np.diff(rho_lin) > 0)
            imax = tmp[-1] + 1
            rspl = InterpolatedUnivariateSpline(rho_lin[:imax], \
                   line_r[i, k, :imax], k=2) 
            
            R[jt, k] = rspl(rho[jt].flatten()).reshape(rho[jt].shape)

            zspl = InterpolatedUnivariateSpline( \
                   rho_line[i, k, :imax], line_z[i, k, :imax], k=2)
            z[jt, k] = zspl(rho[jt].flatten()).reshape(rho[jt].shape)

    return R, z


def mag_theta_star(eqm, t_in, n_rho=400, n_theta=200, rz_grid=False ):
    
    """
    Computes theta star 

    Input:
    ----------
    eqm: equilibrium object
    t_in: float 
        time point for the evaluation
    n_rho: int
        number of flux surfaces equaly spaced from 0 to 1 of rho_pol
    n_theta: int
        number of poloidal points 
    rz_grid: bool
        evaluate theta star on the grid

    Output:
    ----------
    R, z, theta: 3d arrays size(n_rho, n_theta)
    """

    rho = np.linspace(0, 1, n_rho+1)[1:]
    theta = np.linspace(0, 2*np.pi, n_theta, endpoint=False)

    magr, magz = rhoTheta2rz(eqm, rho, theta, t_in=t_in, coord_in='rho_pol')
    magr, magz = magr[0].T, magz[0].T
    
    r0 = np.interp(t_in, eqm.time, eqm.Rmag)
    z0 = np.interp(t_in, eqm.time, eqm.Zmag)

    drdrho, drtheta = np.gradient(magr)
    dzdrho, dztheta = np.gradient(magz)
    dpsidrho, dpsitheta = np.gradient(np.tile(rho**2, (n_theta, 1)).T )

    grad_rho = np.dstack((drdrho, dzdrho, dpsidrho ))
    grad_theta = np.dstack((drtheta, dztheta, dpsitheta))
    normal = np.cross(grad_rho, grad_theta, axis=-1)

    dpsi_dr = -normal[:, :, 0]/(normal[:, :, 2] + 1e-8) #Bz
    dpsi_dz = -normal[:, :, 1]/(normal[:, :, 2] + 1e-8) #Br

#WARNING not defined on the magnetics axis

    dtheta_star = ((magr - r0)**2 + (magz - z0)**2)/(dpsi_dz*(magz - z0) + dpsi_dr*(magr - r0))/magr
    theta = np.arctan2(magz - z0, - magr + r0)
    
    theta = np.unwrap(theta - theta[:, (0, )], axis=1)
    
    from scipy.integrate import cumtrapz

# Definition of the theta star by integral
    theta_star = cumtrapz(dtheta_star, theta, axis=1, initial=0)
    correction = (n_theta - 1.)/n_theta

    theta_star/= theta_star[:, (-1, )]/(2*np.pi)/correction  #normalize to 2pi

    if not rz_grid:
        return magr, magz, theta_star

# Interpolate theta star on a regular grid 
    cos_th, sin_th = np.cos(theta_star), np.sin(theta_star)
    Linterp = LinearNDInterpolator(np.c_[magr.ravel(),magz.ravel()], cos_th.ravel(),0)
         
    nx = 100
    ny = 150

    rgrid = np.linspace(magr.min(), magr.max(), nx)
    zgrid = np.linspace(magz.min(), magz.max(), ny)

    R,Z = np.meshgrid(rgrid, zgrid)
    cos_grid = Linterp(np.c_[R.ravel(), Z.ravel()]).reshape(R.shape)
    Linterp.values[:, 0] = sin_th.ravel() #trick save a some  computing time
    sin_grid = Linterp(np.c_[R.ravel(), Z.ravel()]).reshape(R.shape)  

    theta_star = np.arctan2(sin_grid, cos_grid)

    return rgrid, zgrid, theta_star


def get_q_surf(eqm, qvalue=1., t_in=None, coord_out='rho_tor'):
    """
    Computes the most external magnetic-surface with a given q-value
    If the value is not found, it returns np.nan at that time point

    Input:
    -----------
    eqm: equilibrium object
    qvalue: float > 0
        Desired q-value for magnetic surface
    t_in : float or 1darray or None
        time (eqm.time if None)
    coord_out: str ['rho_pol', 'rho_tor' ,'rho_V', 'r_V', 'Psi','r_a']
        output coordinate label

    Output:
    -----------
    Float or list rho_surf, same size as t_in
    """

    qvalue = float(qvalue)
    if t_in is None:
        t_in = eqm.time
    tarr = np.atleast_1d(t_in)
    unique_idx, index_idx = get_nearest_index(eqm.time, tarr)

    rho_surf = np.zeros_like(tarr)
    rho_q = rho2rho(eqm, eqm.pfl[unique_idx], t_in=eqm.time[unique_idx], coord_in='Psi', coord_out=coord_out)
    qprof = np.abs(eqm.q)
    for ienum, i in enumerate(unique_idx):
        (nrho, ) = eqm.lpfp[i]
        jmin = np.argmin(qprof[i, :nrho])
        q_surf = np.interp(qvalue, qprof[i, jmin:nrho], rho_q[ienum, jmin:nrho], left=np.nan, right=np.nan)
        for j in index_idx[i]:
            rho_surf[j] = q_surf

    return np.squeeze(rho_surf)


def to_geqdsk(equ, t_in=0., cocos_out=1):
    """
    Parameters
    ----------
    eqm: instance of aug_sfutils.sf2equ.EQU
    t_in : float
        Time in seconds.

    Returns
    -------
    dict :
        Dictionary containing CLISTE results in OMFIT geqdsk format.
        Note that this dictionary does not contain yet all processed results typical of 
        omfit_classes.omfit_eqdsk.OMFITgeqdsk objects.
    """

    if cocos_out is not None:
        equ.to_coco(cocos_out=cocos_out)

    jt = np.argmin(np.abs(equ.time - t_in))

    psi_in  = equ.pfl[jt]

    geq = {}

    today = datetime.datetime.today()
    date = today.strftime("%d%b%Y")

    geq['CASE'] = np.array(['  CLISTE ', '    ', '   ', ' #' + str(equ.shot), '  %.2fs' % t_in, '  ' + equ.diag], dtype='<U8')

    geq['CASE2'] = 'CLISTE %s  %s   %d    t~%10.5f     ' %(date, equ.diag, equ.shot, t_in)
    
    # separatrix R and Z
    geq['RBBBS'], geq['ZBBBS'] = rho2rz(equ, 0.999, t_in=equ.time[jt], all_lines=False)
    geq['RBBBS'] = geq['RBBBS'][0][0]
    geq['ZBBBS'] = geq['ZBBBS'][0][0]

    # R,Z of limiter structures
    geq['RLIM'] = [1.035, 1.050, 1.090, 1.140, 1.235, 1.340, 1.430, 1.505, 1.630, 1.770, 1.970, 2.110, 2.170, 2.205, 2.210, 2.175, 2.130, 2.010, 1.890, 1.705, 1.645, 1.580, 1.458, 1.325, 1.235, 1.286, 1.280, 1.245, 1.125, 1.060, 1.035]

    geq['ZLIM'] = [0.000, 0.250, 0.500, 0.700, 0.965, 1.100, 1.150, 1.190, 1.150, 1.085, 0.780, 0.540, 0.390, 0.210, 0.000, -0.150, -0.280, -0.500, -0.680, -0.860, -0.966, -1.208, -1.060, -1.060, -1.126, -0.976, -0.892, -0.820, -0.634, -0.300, 0.000]

    geq['NW'] = len(equ.Rmesh)
    geq['NH'] = len(equ.Zmesh)
    
    geq['RLEFT']   = equ.Rmesh[0]
    geq['RDIM']    = equ.Rmesh[-1] - equ.Rmesh[0]
    geq['ZDIM']    = equ.Zmesh[-1] - equ.Zmesh[0]
    geq['RMAXIS']  = equ.Rmag[jt]
    geq['ZMAXIS']  = equ.Zmag[jt]
    geq['SIMAG']   = psi_in[0]
    geq['SIBRY']   = psi_in[-1]
    geq['CURRENT'] = equ.ip[jt]
    geq['BCENTR']  = equ.B0[jt]
    geq['RCENTR']  = equ.R0  ##[jt]   # TODO this seems to be time-independent?
    geq['ZMID']    = 0.5*(equ.Zmesh[-1] + equ.Zmesh[0])
    geq['PSIRZ'] = equ.pfm[:, :, jt]

    # Interpolate onto an equidistant Psi-grid
    psi_grid = np.linspace(psi_in[0], psi_in[-1], len(equ.Rmesh))

# np.interp fails if "psi_in" is non-monotonic or not ascending
    q  = interp1d(psi_in, equ.q[jt]    , kind='linear', fill_value='extrapolate')
    p  = interp1d(psi_in, equ.pres[jt] , kind='linear', fill_value='extrapolate')
    dp = interp1d(psi_in, equ.dpres[jt], kind='linear', fill_value='extrapolate')
    f  = interp1d(psi_in, 2.e-7*equ.jpol[jt], kind='linear', fill_value='extrapolate')
    df = interp1d(psi_in, -2*np.pi*equ.ffp[jt], kind='linear', fill_value='extrapolate')
    geq['QPSI']   = q(psi_grid)
    geq['PRES']   = p(psi_grid)
    geq['PPRIME'] = dp(psi_grid)
    geq['FPOL']   = f(psi_grid)
    geq['FFPRIM'] = df(psi_grid)

    return geq



import os

def get_diagprefix(name):
    """Determine prefix for path expansion based on static diag list
    The list is copied from gerrit: codac/daq/libddww8/src/islevel0diag.c
    as of 2021-08-20.
    """

    if not isinstance(name, str) or len(name) != 3:
        return None

    # order of dictionary is important due to L0 duplicates !!
    diags = {}

    diags['MH'] = [
        "MHA", "MHB", "MHC", "MHD", "MHE", "MHF", "MHG", "MHH", "MHI"
    ]
    diags['SX'] = [
        "SXA", "SXB", "SXC", "SXD", "SXF", "SXG", "SXH", "SXI", "SXJ", "SXK",
        "SXL", "SXM", "SXN", "SXO", "SXS", "SXT"
    ]
    diags['XX'] = [
        "CGF", "CTA", "CTC", "CXB", "FHA", "FHB", "FHC", "ICA", "ICS", "LIZ",
        "MSX", "NSA", "PCR", "PCS", "PRA", "PRC", "PRF", "PRD", "RFL", "RMA",
        "RMB", "RMC", "TDI", "TSX", "XPR", "XVR", "XVS"
    ]
    diags['YY'] = [
        "CEF", "DTS", "ECQ", "ECE", "ECR", "HEB", "KRF", "LSG", "NPI", "PID",
        "TSC", "UVE", "WPG", "XVT", "XVU"
    ]
    diags['L1'] = [
        "BLB", "BLC", "BPD", "CEC", "CEZ", "COF", "CPZ", "CSX", "DCR", "DCS",
        "DDS", "DTN", "ECK", "ECS", "ECV", "ENR", "EQH", "EQI", "EQR", "FPC",
        "FPG", "FPP", "GIW", "GPI", "GQH", "GQI", "HFC", "ICH", "ICP", "IDA",
        "INJ", "IOB", "JCU", "JFE", "JOU", "JOW", "SPC", "MAE", "MAN", "MAP",
        "MAR", "MBI", "MDO", "MNO", "MOD", "MRI", "MSR", "NEO", "NIS", "NSP",
        "OBS", "RAB", "RAP", "RMD", "RRC", "SCL", "SSX", "TBM", "TFM", "TOT",
        "TTH", "UVS", "VTA", "XRT"
    ]
    diags['L0'] = [
        "ACA", "ACB", "ACQ", "BEP", "BES", "BLK", "BLV", "BOS", "BMT", "BOF",
        "BOL", "BOS", "BRD", "BSK", "CAR", "CCD", "CDH", "CDL", "CER", "CFR",
        "CHR", "CMR", "CMT", "CNR", "COM", "CON", "COO", "COR", "CPR", "CTF",
        "CTI", "CUR", "CXR", "DCD", "DCN", "DEP", "DIN", "DIV", "DSM", "DST",
        "DTM", "DUS", "DVM", "DVS", "ECH", "ECN", "ECU", "END", "EVS", "EVU",
        "EVV", "EZC", "EZD", "FHF", "FVS", "GRA", "GRF", "GRG", "GVS", "HAM",
        "HAR", "HBD", "HDV", "HEL", "HFB", "HST", "HTS", "HVS", "HXR", "IFM",
        "ICF", "ICG", "ICL", "ICR", "ICT", "ION", "JMO", "JOH", "KMT", "KWK",
        "KWN", "KWS", "LBM", "LBO", "LEN", "LIA", "LIB", "LIC", "LIF", "LPS",
        "LSB", "LSF", "LSM", "LSW", "LVS", "MAB", "MAC", "MAD", "MAF", "MAG",
        "MAH", "MAI", "MAK", "MAM", "MAS", "MAU", "MAW", "MAX", "MAY", "MBR",
        "MER", "MGS", "MHA", "MHB", "MHC", "MHD", "MHE", "MHF", "MHG", "MHH",
        "MIC", "MIR", "MPC", "MPG", "MSE", "MSP", "MSS", "MSX", "MUM", "NIB",
        "NIK", "NIR", "NIT", "NPA", "NWB", "NWK", "OSI", "PEL", "PHA", "PHB",
        "PKG", "POT", "PPA", "PPT", "PSL", "RAD", "RAH", "RAV", "REF", "REH",
        "REI", "REV", "RFL", "RMA", "RMB", "ROE", "RSG", "RWB", "SCJ", "SIF",
        "SMF", "SPU", "SST", "TAA", "TAB", "TAC", "TAD", "TAE", "TAF", "TAG",
        "TAH", "TEO", "TER", "TET", "TLS", "UVD", "VEC", "VSS", "VTS", "ZEA",
        "ZEB", "ZEF"
    ]

    for pr in diags.keys():
        if name in diags[pr]:
            return pr

    return 'L1'

def shf(dirlist, i):
    """Check existence of shotfile in the directory given
    """
    for d in dirlist:
        if d.startswith(str(i)):
            return True
    return False

def ddcshotnr(diag, shot=99999, exp='AUGD'):
    """Py-ddcshotnr
    Gets next shotnumber of specified diagnostic from
    specified experiment smaller than the provided shotnumber

    Input:
        diag        str  Diagnostic
        shot(opt)   int  Shot number (default: 99999)
        exp(opt)    str  Exp (default: 'AUGD')
    Error codes:
        -1  :  no suitable shot found under given experiment
        -2  :  wrong input shotnumber (wrong type or out of range)
    """

    # shotnumber needs to be an int and in range, else no result
    if not isinstance(shot, int) or shot > 99999:
        return -3
    # AUGD shotfiles follow a particular path-specification
    if exp == 'AUGD':
        basepath = '/afs/ipp-garching.mpg.de/u/augd/shots/'
        prefix = get_diagprefix(diag)
        fac = 10
    else:
        # private shotfiles use a different path logic
        basepath = '/afs/ipp-garching.mpg.de/u/'+exp.lower()+'/shotfiles/'+diag+'/'
        fac = 1000

    # go backwards starting with directory of given shot
    for sr in range(int(shot/fac), -1, -1):
        fullpath = basepath + '%04d' % sr + '/' + prefix + '/' + diag \
            if exp == 'AUGD' else basepath + str(sr)
        if not os.path.isdir(fullpath):
            continue
        dirlist = os.listdir(fullpath)
        ista = 9 if exp == 'AUGD' else 999
        for i in range(ista, -1, -1):
            # if maxshot happens to be smaller than current check, skip the check
            if shot < sr*fac+i:
                continue
            fshot = '%05d' % int(sr*fac+i) if exp == 'AUGD' else '%03d' % i
            if shf(dirlist, fshot):
                return sr*fac+i

    # no result found anywhere, return -1
    return -1


def previousshot(diag, shot=99999, exp='AUGD'):
    """Alias of ddcshotnr
    """
    return ddcshotnr(diag, shot=shot, exp=exp)


if __name__ == "__main__":
    pass



import logging, traceback
from struct import unpack, pack
import numpy as np
from aug_sfutils import sfdics, str_byt

n_rel = 8

logger = logging.getLogger('aug_sfutils.sfhread')
logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


for key, val in sfdics.obj_name.items():
    exec('%s=%d' %(val, key))


def read_sfh(byt_str):
    """
    Reads a full shotfile header
    """

    sfhead = {}
    obj_names = []
    n_max = 1000
    n_obj = n_max
    for j in range(n_max):
        sfo = SFH_READ(byt_str[j*128: (j+1)*128])
        if hasattr(sfo, 'obj_type'):
            sfo.objid = j
            onam = str_byt.to_str(sfo.objnam.strip())
            if sfo.obj_type == Diagnostic:
                if n_obj == n_max: # There might be several diags objects in a SFH
                    n_obj = sfo.num_objs
            sfhead[onam] = sfo
            obj_names.append(onam)
        if j >= n_obj - 1:
            break

    for sfo in sfhead.values():
        sfo.relations = [obj_names[jid] for jid in sfo.rel if jid != 65535]

    return sfhead


class SFH_READ:
    """
    Reads a generic object's metadata from the SFH's byte string
    """

    def __init__(self, byte_str):
        """
        Reads the SFH part of an object, common to all objects
        """
        self.objnam = str_byt.to_str(byte_str[0:8])

        if not self.objnam:
            logger.error('Error: empty object name')
            return

        self.obj_type, self.level, self.status = unpack('>3H', byte_str[ 8: 14])
        typ = self.obj_type
        if typ in sfdics.obj_name.keys():
            self.object_type = sfdics.obj_name[typ]
        else:
            self.object_type = 'Unknown'
        self.errcode  = unpack('>h', byte_str[14: 16])[0]
        fmt = '>%dH' %n_rel
        self.rel      = list(unpack(fmt,  byte_str[16        : 16+2*n_rel]))
        self.address, self.length = unpack('>2I', byte_str[16+2*n_rel: 24+2*n_rel])
        self.val      = byte_str[40:  64]
        logger.debug(self.objnam)
        self.descr    = str_byt.to_str(byte_str[64: 128].strip())
        logger.debug(self.descr)

        val_func = { \
            1 : self.sfdiag, 2: self.list, 3: self.device, 4: self.parmset, \
            5 : self.mapping, 6: self.sig, 7: self.sig, 8: self.tbase, \
            9 : self.sf_list, 10: self.algorithm, 11: self.update_set, \
            12: self.loctimer, 13: self.abase, 14: self.qualifier, \
            15: self.modobj, 16: self.mapext, 17: self.resource, \
            18: self.addrlen }

        if typ in val_func.keys():
            val_func[typ]()
        else:
            logger.error('Object type %d of object %s not supported', self.obj_type, self.objnam)


    def sfdiag(self):
        """
        Metadata of a DIAG object
        """
        self.diag_code = self.val[: 4]
        self.num_objs, self.diag_type  = unpack('>2H', self.val[4: 8])
        self.shot_nr , self.c_time     = unpack('>2I', self.val[8: 16])
        self.up_limit, self.exp, self.version, self.s_type = unpack('>4H', self.val[16: 24])


    def list(self):
        """
        Metadata of a LIST object
        """

        self.data_format, self.nitems, self.ordering, self.list_type = unpack('>4H', self.val[ : 8])


    def device(self):
        """
        Metadata of a DEVICE object
        """

        self.data_format, self.acquseq, self.nitems, self.dev_type = unpack('>4H', self.val[:8])
        self.dev_addr, self.n_chan  = unpack('>2I', self.val[ 8: 16])
        self.task    , self.dev_num = unpack('>2H', self.val[16: 20])
        self.n_steps                = unpack('>I' , self.val[20: 24])[0]  


    def parmset(self):
        """
        Metadata of a ParameterSet object
        """

        self.nitems, self.cal_type = unpack('>2H', self.val[4: 8])
        if self.cal_type in sfdics.cal_type.keys():
            self.calibration_type = sfdics.cal_type[self.cal_type]


    def mapping(self):
        """
        Metadata of a MAP object
        """

        self.nitems, self.map_type = unpack('>2H' , self.val[4: 8])
        self.task = unpack('>H' , self.val[16: 18])[0]


    def sig(self):
        """
        Metadata of a Signal or SignalGroup object
        """

        self.data_format, self.physunit, self.num_dims = unpack('>3H' , self.val[: 6])
        if self.physunit in sfdics.unit_d.keys():
            self.phys_unit = sfdics.unit_d[self.physunit]
        else:
            logger.warning('No phys. unit found for object %s, key=%d', self.objnam, self.physunit)
            self.phys_unit = ''
        self.stat_ext = unpack('>h' , self.val[6: 8])[0]
        self.index    = list(unpack('>4I', self.val[8: 24]))


    def tbase(self):
        """
        Metadata of a TIMEBASE object
        """

        self.data_format, self.burstcount, self.event, self.tbase_type = unpack('>4H', self.val[: 8])
        self.s_rate = unpack('>I', self.val[ 8: 12])[0] #Hz
        self.n_pre, self.n_steps = unpack('>2I', self.val[16: 24])
        if self.tbase_type in sfdics.tbtype_d.keys():
            self.timebase_type = sfdics.tbtype_d[self.tbase_type]


    def sf_list(self):
        """
        Metadata of a SF_LIST object
        """

        self.nitems = unpack('>H', self.val[2: 4])


    def algorithm(self):
        """
        Metadata of an ALGORITHM object
        """

        self.hostname = self.val[ 8: 16]
        self.date     = self.val[16: 24] 


    def update_set(self):
        """
        Metadata of an UPDATE_SET object
        """

        self.nitems =  unpack('>H' , self.val[ 2: 4])[0]
        self.input_vals = unpack('>i' , self.val[ 4: 8])[0]
        self.next_index, self.size = unpack('>I' , self.val[ 16: 24])


    def loctimer(self):
        """
        Metadata of an LOCTIMER object
        """

        self.data_format, self.resolution = unpack('>2H', self.val[: 4])
        self.size = unpack('>I', self.val[20: 24])[0]


    def abase(self):
        """
        Metadata of an AREABASE object
        """

        self.data_format = unpack('>H' , self.val[ : 2])[0]
        self.physunit    = unpack('>3H', self.val[2: 8])
        self.phys_unit = [sfdics.unit_d[x] for x in self.physunit]
        self.size_x, self.size_y, self.size_z, self.n_steps = unpack('>4I' , self.val[8: 24])
        self.sizes = [self.size_x, self.size_y, self.size_z]


    def qualifier(self):
        """
        Metadata of a QUALIFIER object
        """

        self.data_format = unpack('>H' , self.val[ : 2])[0]
        self.num_dims, self.qsub_typ = unpack('>2H' , self.val[4: 8])
        self.index_4, self.index_3, self.index_2, self.max_sections = unpack('>4I' , self.val[8: 24])


    def modobj(self):
        """
        Metadata of a MODOBJ object
        """

        self.nitems = unpack('>H' , self.val[ : 2])[0]


    def mapext(self):
        """
        Metadata of a MAPEXT object
        """

        self.nitems, self.mapalg = unpack('>2H' , self.val[ : 4])
        self.tbeg  , self.tend   = unpack('>2I', self.val[4: 12])
        self.val_0 , self.val_1  = unpack('>2H', self.val[12: 16])
        self.val_2 , self.val_3  = unpack('>2I', self.val[16: 24])


    def resource(self):
        """
        Metadata of a RESOURCE object
        """

        self.num_cpus, self.first_cpu = unpack('>2H' , self.val[ : 4])


    def addrlen(self):
        """
        Value of ADDRLEN object
        """

        self.addrlen = unpack('>H' , self.val[ : 2])[0]



import os, struct, logging
import numpy as np

if os.name == 'nt':
    afs_home = '\/afs/ipp-garching.mpg.de/home'
else:
    afs_home = '/afs/ipp-garching.mpg.de/home'

logger = logging.getLogger('aug_sfutils.manage_ed')
logger.level = logging.INFO


def sf_path(nshot, diag, exp='AUGD', ed=0):
    """Path finder for shotfiles

    Input:
        nshot    int  Shotnumber
        diag     str  Diagnostic label
        exp(opt) str  Expriment label (default: 'AUGD')
        ed(opt)  int  Edition number (default: 0)

    Output:
        path_out   str   Full path of shotfile (None if not found)
        ed         int   Actual edition (non trivial if input ed=0)
    """

    sshot = str(nshot).zfill(5)
    exp = exp.lower()
    diag = diag.upper()

    if ed == 0:
        ed = ed_zero(diag, nshot, exp=exp)
    if ed is None:
        return None, None

    if exp == 'augd':
        sf_dir = '%s/a/augd/shots/%s' %(afs_home, sshot[:4])
        a = os.listdir(sf_dir)
        path_out = None
        for subdir in a:
            path1 = '%s/%s/%s' %(sf_dir, subdir, diag)
            if os.path.isdir(path1):
                if not os.path.isfile('%s/ed_cntl' %path1):
                    path_out = '%s/%05d' %(path1, nshot)
                else:
                    path_out = '%s/%d.%d' %(path1, nshot, ed)
                break
    else:
        path1 = '%s/%s/%s/shotfiles/%s/%s' %(afs_home, exp[0], exp, diag, sshot[:2])
        if os.path.isdir(path1):
            if not os.path.isfile('%s/ed_cntl' %path1):
                path_out = '%s/%s' %(path1, sshot[2:])
            else:
                path_out = '%s/%s.%d' %(path1, sshot[2:], ed)

    return path_out, ed


def ed_zero(diag, nshot, exp='augd'):
    """Finds the actual edition number of ed=0
    Input:
        nshot    int  Shotnumber
        diag     str  Diagnostic label
        exp(opt) str  Expriment label (default: 'AUGD')

    Output:
        ed_nr    int  Actual edition number. 1 for level0, None if not found
    """
 
    sshot = str(nshot).zfill(5)
    exp = exp.lower()
    diag = diag.upper()

    if exp == 'augd':
        sf_dir = '%s/a/augd/shots/%s' %(afs_home, sshot[:4])
        if not os.path.isdir(sf_dir):
            return None
        a = os.listdir(sf_dir)
        fed_dir = None
        for subdir in a:
            fed_dir = '%s/%s/%s' %(sf_dir, subdir, diag)
            if os.path.isdir(fed_dir):
                if not os.path.isfile('%s/ed_cntl' %fed_dir):
                    return 1
                break
    else:
        fed_dir = '%s/%s/%s/shotfiles/%s/%s' %(afs_home, exp[0], exp, diag, sshot[:2])
        nshot = nshot % 1000
        if not os.path.isdir(fed_dir):
            return None
        else:
            if not os.path.isfile('%s/ed_cntl' %fed_dir):
                return 1

    ed_d = read_ed_cntl(fed_dir, exp=exp)
    if ed_d is None:
        return None
    logger.debug(nshot)
    if nshot in ed_d.keys():
        if ed_d[nshot] == -1:
            logger.error('Edition control found ed=-1')
            return None
        else:
            return ed_d[nshot]
    else:
        return None


def read_ed_cntl(fed_dir, exp='augd'):
    """Parser of ed_cntl file

    Input:
        fed_dir   str   Path where to look for ed_cntl file
        exp(opt)  str   Exp
    Output:
        max_ed    int   Max edition number for given diag, shot
    """
    ed_ctrl = '%s/ed_cntl' %fed_dir
    if not os.path.isfile(ed_ctrl):
        return None

    exp = exp.strip().lower()
    if exp == 'augd':
        shot_byt  = 5
        delta_byt = 24
    else:
        shot_byt  = 3
        delta_byt = 20

    with open(ed_ctrl, 'rb') as f:
        byt_str = f.read()

    jbyt = 12
    max_ed = {}

    while(True):
        try:
            ed = struct.unpack('>I', byt_str[jbyt + 4: jbyt + 8])[0]
            shot = struct.unpack('>%dc' %shot_byt, byt_str[jbyt+16: jbyt+16+shot_byt])
            jbyt += delta_byt
            sshot = b''.join(shot)
            nshot = int(sshot)
            max_ed[nshot] = np.int32(ed)
        except:
            break

    if len(max_ed.keys()) > 0:
        return max_ed
    else:
        return None


if __name__ == '__main__':

    fed_in = '%s/g/git/shotfiles/TRA/35' %afs_home
    ed = read_ed_cntl(fed_in, exp='git')
    logger.info(fed_in, ed) 
    fed_in = '%s/a/augd/shots/3956/L1/NSP' %afs_home
    fed_in = '%s/a/augd/shots/3956/L1/TOT' %afs_home
    ed = read_ed_cntl(fed_in)
    logger.info(fed_in, ed)



import logging
import numpy as np
from scipy import interpolate
from aug_sfutils import time_no_elm

logger = logging.getLogger('trview.time_slice')
logger.setLevel(logging.INFO)


def map2tgrid_rho_err(raw_d, t_out, nshot=None, noELMs=False, dt_out=None):

# t_out is the output regular time grid

    t1 = raw_d['tgrid']
    nt_in, nx_in = raw_d['rhop'].shape
    nt_out = len(t_out)

    if nt_in == 1:
        x_out   = np.tile(raw_d['rhop'    ][0, :], nt_out).reshape((nt_out, nx_in)) 
        y_out   = np.tile(raw_d['data'    ][0, :], nt_out).reshape((nt_out, nx_in)) 
        err_out = np.tile(raw_d['data_err'][0, :], nt_out).reshape((nt_out, nx_in)) 
        return x_out, y_out, err_out

# Remove data close to ELMs

    if (nshot is not None) and noELMs:
        ind_elm = time_no_elm.time_no_elm(nshot, t1) # Time points too close to ELMs
        ind_ok = (~ind_elm)
    else:
        ind_ok = np.ones(nt_in, dtype=bool)

    t_in   = raw_d['tgrid'   ][ind_ok]
    x_in   = raw_d['rhop'    ][ind_ok, :]
    y_in   = raw_d['data'    ][ind_ok, :]
    err_in = raw_d['data_err'][ind_ok, :]

    dt_in = np.min(np.diff(t_in))

# Remove infinite/nan from measurement uncertainties

    ymax = np.nanmax(y_in)
    non_fin_y   = (~np.isfinite(y_in))
    non_fin_err = (~np.isfinite(err_in))
    y_in[non_fin_y] = 0
    y_in[non_fin_err] = 0
    err_in[non_fin_y] = np.nan #1e3*ymax
    err_in[non_fin_err] = np.nan #1e3*ymax

# git 23.04.21
    ind_zero = (y_in == 0)
    y_in[ind_zero] = np.nan 
    err_in[ind_zero] = np.nan

# In case of no valid points in time interval

    fx = interpolate.interp1d(t_in, x_in, axis=0)
    fy = interpolate.interp1d(t_in, y_in, axis=0)
    fe = interpolate.interp1d(t_in, err_in, axis=0)

# Create intermediate grid
    if dt_out is None:
        dt_out = t_out[-1] - t_out[-2]

    x_out   = np.zeros((nt_out, nx_in))
    y_out   = np.zeros((nt_out, nx_in))
    err_out = np.zeros((nt_out, nx_in))

    tbins = np.zeros(nt_out + 1)
    tbins[0] = t_out[0] - 0.5*dt_out
    tbins[1: -1] = 0.5*(t_out[1:] + t_out[:-1])
    tbins[-1] = t_out[-1] + 0.5*dt_out
    indbin = np.searchsorted(tbins, t_in, side='left')
    for jt in range(nt_out):
        (tind, ) = np.where(indbin == jt+1)
        if np.isnan(y_in[tind, :]).all():
            logger.debug(' All nan at t=%.3f', t_out[jt])
        if tind.any():
            x_out[jt]   = np.nanmean(x_in[tind, :], axis=0)
            y_out[jt]   = np.nanmean(y_in[tind, :], axis=0)
#            err_out[jt] = np.linalg.norm(err_in[tind, :], axis=0)/len(tind)
            err_out[jt] = np.nanmean(err_in[tind, :], axis=0)
        else:
            tloc = t_out[jt]
            if (tloc >= t_in[0]) and (tloc <= t_in[-1]): #interpolate internal points
                x_out  [jt] = fx(tloc)
                y_out  [jt] = fy(tloc)
                err_out[jt] = fe(tloc)
            else:
                if (tloc < t_in[0] - dt_in): # set to 0 beyond dt
                    x_out[jt] = x_in[0]
                    y_out[jt] = 0
                    err_out[jt] = 1e3*ymax
                elif tloc < t_in[0]:      # extrapolate within dt
                    tfac = (tloc - t_in[0])/(t_in[1] - t_in[0])
                    x_out[jt] = x_in[0] + tfac*(x_in[1] - x_in[0])
                    y_out[jt] = y_in[0] + tfac*(y_in[1] - y_in[0])
                    err_out[jt] = err_in[0]
                if (tloc > t_in[-1] + dt_in) :
                    x_out[jt] = x_in[-1]
                    y_out[jt] = 0
                    err_out[jt] = 1e3*ymax
                elif tloc > t_in[-1]:
                    tfac = (tloc - t_in[-1])/(t_in[-2] - t_in[-1])
                    x_out[jt] = x_in[-1] + tfac*(x_in[-2] - x_in[-1])
                    y_out[jt] = y_in[-1] + tfac*(y_in[-2] - y_in[-1])
                    err_out[jt] = err_in[-1]

    err_out[np.isnan(err_out)] = 1e3*ymax

    return x_out, y_out, err_out



def map2tgrid(t_out, t_in, sig_input, nshot=None, noELMs=False, ind_elm=None, indbin=None):

# Remove data close to ELMs

    if ind_elm is None:
        if (nshot is not None) and noELMs:
            print('Discard times close to ELMs')
            ind_elm = time_no_elm.time_no_elm(nshot, t_in) # Time points too close to ELMs
            logger.info('#ELM discarded: %d out of %d', np.sum(ind_elm), len(t_in))
        else:
            ind_elm = []

    dt_left  = t_in[ 1] - t_in[ 0]
    dt_right = t_in[-1] - t_in[-2]
    nt_out = len(t_out)

# Create intermediate grid
# Output grid must be regular!

    if indbin is None:
        dt_out = t_out[-1] - t_out[-2]
        tbins = np.zeros(nt_out + 1)
        tbins[ 0] = t_out[ 0] - 0.5*dt_out
        tbins[-1] = t_out[-1] + 0.5*dt_out
        tbins[1: -1] = 0.5*(t_out[1:] + t_out[:-1])
        indbin = np.searchsorted(tbins, t_in, side='left')

# Back-up for empty slices

    sig_in = sig_input.copy()
    fy = interpolate.interp1d(t_in, sig_in, axis=0)
    sig_in[ind_elm, ...] = np.nan

    if sig_in.ndim == 1:
        sig_out = np.zeros(nt_out)
    else:
        dims = np.append(nt_out, sig_in.shape[1:])
        sig_out = np.zeros(dims)

    for jt in range(nt_out):
        (tind,) = np.where(indbin == jt+1)
        if tind.any():
            sig_out[jt] = np.nanmean(sig_in[tind], axis=0)
        else:
            tloc = t_out[jt]
            if (tloc >= t_in[0]) and (tloc <= t_in[-1]): #interpolate internal points
                sig_out[jt] = fy(tloc)
            else:
                if (tloc < t_in[0] - dt_left): # set to 0 beyond dt
                    sig_out[jt] = 0
                elif tloc < t_in[0]:      # extrapolate within dt
                    tfac = (tloc - t_in[0])/(t_in[1] - t_in[0])
                    sig_out[jt] = sig_in[0] + tfac*(sig_in[1] - sig_in[0])
                if (tloc > t_in[-1] + dt_right) : # set to 0 beyond dt
                    sig_out[jt] = 0
                elif tloc > t_in[-1]:      # extrapolate within dt
                    tfac = (tloc - t_in[-1])/(t_in[-2] - t_in[-1])
                    sig_out[jt] = sig_in[-1] + tfac*(sig_in[-2] - sig_in[-1])

    return sig_out, ind_elm, indbin



#!/usr/bin/env python

"""Shotfile reading with pure python

https://www.aug.ipp.mpg.de/aug/manuals/aug_sfutils

"""
__author__  = 'Giovanni Tardini (Tel. 1898)'
__version__ = '0.8.0'
__date__    = '29.11.2022'

import os, sys, logging, traceback

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%H:%M:%S')

logger = logging.getLogger('aug_sfutils')

if len(logger.handlers) == 0:
    hnd = logging.StreamHandler()
    hnd.setFormatter(fmt)
    logger.addHandler(hnd)

#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)

logger.info('Using version %s', __version__)

try: # wrapper classes, available only with afs-client and kerberos access
    from .ww import *
except:
    traceback.print_exc()
    logger.warning('ww not loaded')
try:
    from .sfh import *
except:
    traceback.print_exc()
    logger.warning('sfh not loaded')

from .sfread import *
from .sf2equ import *
from .libddc import ddcshotnr, previousshot
from .mapeq import *
from .getlastshot import getlastshot
sf_home = os.path.dirname(os.path.realpath(__file__))
logger.info('AUG-SF home %s', sf_home)



import logging
import aug_sfutils as sf
from aug_sfutils import sfdics, sfhwrite

fmt = logging.Formatter('%(asctime)s | %(name)s | %(levelname)s: %(message)s', '%H:%M:%S')
hnd = logging.StreamHandler()
hnd.setFormatter(fmt)
logger = logging.getLogger('sfhmod')
logger.addHandler(hnd)
#logger.setLevel(logging.DEBUG)
#logger.setLevel(logging.INFO)


for key, val in sfdics.obj_name.items():
    exec('%s=%d' %(val, key))


class SFHMOD:


    def __init__(self, fin=None):

        self.sfo = sf.SFREAD(sfh=fin)


    def modtime(self, name, nt):
        '''Changes the size of the time dimension of a TB, AB, Sig or SigGroup (if related to a TB)'''

        sfobj = self.sfo.sfh[name]
        otyp = sfobj.obj_type
        if name not in self.sfo.objects:
            logger.error('modtim accepts only TB, AB, Sig, SigName')
        elif otyp in (TimeBase, AreaBase):
            sfobj.n_steps = nt
        elif otyp == Signal:
            sfobj.index[-1] = nt
        elif otyp == SignalGroup:
            tdim = None
            for jrel, rel in enumerate(sfobj.relations):
                if self.sfo.sfh[rel].obj_type == TimeBase:
                    tdim = jrel
                    break
            if tdim is not None:
                sfobj.index[3-tdim] = nt


    def modtimeall(self, tbase, nt):
        '''Changes the size of the time dimension of input TB plus all AB, Sig or SigGroup related to it'''

        sfobj = self.sfo.sfh[tbase]
        if sfobj.obj_type not in (TimeBase, Signal):
            logger.error('modtimall needs a TB or SIG name')
            return

        self.modtime(tbase, nt)

        for sfobj in self.sfo.sfh.values():
            onam = sfobj.objnam
            if onam in self.sfo.objects:
                for rel in sfobj.relations:
                    if rel == tbase:
                        self.modtime(onam, nt)


    def modindex(self, sgr, index):
        '''Changes the dims of a SigGroup'''

        index = list(index)
        sfobj = self.sfo.sfh[sgr]
        if sfobj.obj_type != SignalGroup:
            logger.error('modindex needs a SigGroup')
            return
        while len(index) < 4:
            index.append(1)
        sfobj.index = index[::-1]


    def modareasize(self, abase, size_x=None, size_y=None, size_z=None):
        '''Changes the sizes of an AreaBase'''

        sfobj = self.sfo.sfh[abase]
        if sfobj.obj_type != AreaBase:
            logger.error('modareasize needs an AreaBase')
            return

        if size_x is not None:
            sfobj.size_x = size_x
        if size_y is not None:
            sfobj.size_y = size_y
        if size_z is not None:
            sfobj.size_z = size_z


    def modareaall(self, abase, nx):
        '''Changes size_x of a space-1D AreaBase and the spatial dim of all related SigGroups'''
 
        sfobj = self.sfo.sfh[abase]
        if sfobj.obj_type != AreaBase:
            logger.error('modareaall needs an AreaBase')
            return
        if sfobj.size_y > 0:
            logger.error('modareaall works only with space-1d AreaBases')
            return

        sfobj.size_x = nx

        for sfobj in self.sfo.sfh.values():
            otyp = sfobj.obj_type
            if otyp == SignalGroup:
                adim = None
                for jrel, rel in enumerate(sfobj.relations):
                    if rel == abase:
                        adim = jrel
                        break
                if adim is not None:
                    sfobj.index[3-adim] = nx


    def modpar(self, pset, pnam, data):
        '''Modify data content (and length) of parameter pnaf of ParSet pset'''

        par = self.sfo.sfh[pset].pars[pnam]
        par.n_items = len(data)
        par[:par.n_items] = data[:]


    def write(self, fout='mytest.fsh'):

        sfhwrite.write_sfh(self.sfo, fout)


if __name__ == '__main__':

    fsfh = 'RAB00000.sfh'
    sfh = SFHMOD(fin=fsfh)
    sfh.modtimeall('time', 12)
    print('')
    sfh.modareaall('rho_in', 21)
    print('')
    sfh.write(fout='myrab.sfh')

    fsfh = 'TST00000.sfh'
    sfh = SFHMOD(fin=fsfh)    
    sfh.modpar('NBIpar', 'rtcena', [0.,1.,2.,3.] )
    print('')
    sfh.write(fout='mytra.sfh')



import logging
import numpy as np
import aug_sfutils as sf

logger = logging.getLogger('trview.time_no_elm')


def time_no_elm(nshot, t_in, dt=[0.001, 0.005]):

    tloc = t_in
    ldel = np.zeros(len(t_in), dtype=bool)

    elm = sf.SFREAD(nshot, 'ELM')
    if elm.status:
        t_elms = elm.getobject('tELM')
        if t_elms is None:
            logger.error('No time frames discarded close to ELMs')
            return ldel
#elm synchronized equlibrium - it will not use timepoits during the elms, CLISTE is not working well
        if np.max(t_elms > 0):
            told = -np.infty
            for it, t in enumerate(tloc):
                if any(t_elms<t):
                    last = np.argmax(t_elms[t_elms<t])
                    last_elm = t_elms[last]
                    if last < len(t_elms)-1:
                        next_elm = t_elms[last+1]
# Remove points up to 5ms after an ELM crash or 1 ms before
                        if t <= last_elm + dt[1] or t >= next_elm - dt[0] :
                            ldel[it] = True
                    else:
                        if t <= last_elm + dt[1]:
                            ldel[it] = True
                told = tloc[it-1]
        else:
            logger.error('No ELM correction')
    else:
        logger.error('No ELM correction')

    return ldel
